"""JSONL bridge for forwarding events between runtimes.

This bridge is intentionally simple:
- emit/dispatch appends one raw event JSON object per line
- listener polls the file and emits any unseen lines
"""

from __future__ import annotations

import asyncio
import json
from collections.abc import Callable
from pathlib import Path
from typing import Any

from uuid_extensions import uuid7str

from abxbus.base_event import BaseEvent
from abxbus.event_bus import EventBus, EventPatternType, in_handler_context


class JSONLEventBridge:
    def __init__(self, path: str, *, poll_interval: float = 0.25, name: str | None = None):
        self.path = Path(path)
        self.poll_interval = poll_interval
        self._inbound_bus = EventBus(name=name or f'JSONLEventBridge_{uuid7str()[-8:]}', max_history_size=0)

        self._running = False
        self._start_task: asyncio.Task[None] | None = None
        self._start_lock = asyncio.Lock()
        self._listener_task: asyncio.Task[None] | None = None
        self._byte_offset = 0
        self._pending_line = ''

    def on(self, event_pattern: EventPatternType, handler: Callable[[BaseEvent[Any]], Any]) -> None:
        self._ensure_started()
        self._inbound_bus.on(event_pattern, handler)

    async def emit(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        self._ensure_started()

        payload = event.model_dump(mode='json')
        self.path.parent.mkdir(parents=True, exist_ok=True)

        await asyncio.to_thread(self._append_line, json.dumps(payload, separators=(',', ':')))

        if in_handler_context():
            return None
        return event

    async def dispatch(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        return await self.emit(event)

    async def start(self) -> None:
        current_task = asyncio.current_task()
        if self._start_task is not None and self._start_task is not current_task and not self._start_task.done():
            await self._start_task
            return

        if self._running:
            return

        try:
            async with self._start_lock:
                if self._running:
                    return
                self.path.parent.mkdir(parents=True, exist_ok=True)
                self.path.touch(exist_ok=True)
                self._byte_offset = self.path.stat().st_size
                self._pending_line = ''
                self._running = True
                if self._listener_task is None or self._listener_task.done():
                    self._listener_task = asyncio.create_task(self._listen_loop())
        finally:
            if self._start_task is current_task:
                self._start_task = None

    async def close(self, *, clear: bool = True) -> None:
        if self._start_task is not None:
            self._start_task.cancel()
            await asyncio.gather(self._start_task, return_exceptions=True)
            self._start_task = None
        self._running = False
        if self._listener_task is not None:
            self._listener_task.cancel()
            await asyncio.gather(self._listener_task, return_exceptions=True)
            self._listener_task = None
        await self._inbound_bus.destroy(clear=clear)

    def _ensure_started(self) -> None:
        if self._running:
            return
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return
        if self._start_task is None or self._start_task.done():
            self._start_task = asyncio.create_task(self.start())

    async def _listen_loop(self) -> None:
        while self._running:
            try:
                await self._poll_new_lines()
            except asyncio.CancelledError:
                raise
            except Exception:
                pass
            await asyncio.sleep(self.poll_interval)

    async def _poll_new_lines(self) -> None:
        previous_offset = self._byte_offset
        appended_text, new_offset = await asyncio.to_thread(self._read_appended_text, previous_offset)
        self._byte_offset = new_offset

        if new_offset < previous_offset:
            self._pending_line = ''

        if not appended_text:
            return

        combined_text = self._pending_line + appended_text
        new_lines = combined_text.split('\n')
        self._pending_line = new_lines.pop() if new_lines else ''

        for line in new_lines:
            line = line.strip()
            if not line:
                continue
            try:
                payload = json.loads(line)
            except Exception:
                continue
            await self._dispatch_inbound_payload(payload)

    async def _dispatch_inbound_payload(self, payload: Any) -> None:
        event = BaseEvent[Any].model_validate(payload).event_reset()
        self._inbound_bus.emit(event)

    def _read_appended_text(self, offset: int) -> tuple[str, int]:
        try:
            with self.path.open('r', encoding='utf-8') as fp:
                fp.seek(0, 2)
                file_size = fp.tell()

                start_offset = 0 if file_size < offset else offset
                if file_size == start_offset:
                    return '', file_size

                fp.seek(start_offset)
                return fp.read(), fp.tell()
        except FileNotFoundError:
            return '', 0

    def _append_line(self, payload: str) -> None:
        with self.path.open('a', encoding='utf-8') as fp:
            fp.write(payload + '\n')



"""PostgreSQL LISTEN/NOTIFY + flat-table bridge for forwarding events.

Optional dependency: asyncpg

Connection URL format:
    postgresql://user:pass@host:5432/dbname[/tablename]?sslmode=require

Schema shape:
- event_id (PRIMARY KEY)
- event_created_at (indexed)
- event_type (indexed)
- event_payload (full event JSON payload)
- one TEXT column per event_* field storing JSON-serialized values
"""

from __future__ import annotations

import asyncio
import importlib
import json
import re
from collections.abc import Callable
from typing import Any, TypeGuard
from urllib.parse import urlsplit, urlunsplit

from uuid_extensions import uuid7str

from abxbus.base_event import BaseEvent
from abxbus.event_bus import EventBus, EventPatternType, in_handler_context

_IDENTIFIER_RE = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')
_DEFAULT_POSTGRES_TABLE = 'abxbus_events'
_DEFAULT_POSTGRES_CHANNEL = 'abxbus_events'
_EVENT_PAYLOAD_COLUMN = 'event_payload'


def _validate_identifier(identifier: str, *, label: str) -> str:
    if not _IDENTIFIER_RE.match(identifier):
        raise ValueError(f'Invalid {label}: {identifier!r}. Use only [A-Za-z0-9_] and start with a letter/_')
    return identifier


def _parse_table_url(table_url: str) -> tuple[str, str]:
    parsed = urlsplit(table_url)
    segments = [segment for segment in parsed.path.split('/') if segment]
    if len(segments) < 1:
        raise ValueError(
            'PostgresEventBridge URL must include at least database in path, e.g. '
            'postgresql://user:pass@host:5432/dbname[/tablename]'
        )

    db_name = segments[0]
    table_name = _validate_identifier(segments[1], label='table name') if len(segments) >= 2 else _DEFAULT_POSTGRES_TABLE

    dsn_path = f'/{db_name}'
    dsn = urlunsplit((parsed.scheme, parsed.netloc, dsn_path, parsed.query, parsed.fragment))
    return dsn, table_name


def _index_name(table: str, suffix: str) -> str:
    return _validate_identifier(f'{table}_{suffix}'[:63], label='index name')


def _is_str_keyed_dict(value: Any) -> TypeGuard[dict[str, Any]]:
    return isinstance(value, dict)


def _split_bridge_payload(payload: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any]]:
    event_fields: dict[str, Any] = {}
    event_payload: dict[str, Any] = dict(payload)
    for key, value in payload.items():
        if key.startswith('event_'):
            event_fields[key] = value
    return event_fields, event_payload


class PostgresEventBridge:
    def __init__(self, table_url: str, channel: str | None = None, *, name: str | None = None):
        self.table_url = table_url
        self.dsn, self.table = _parse_table_url(table_url)
        derived_channel = channel or _DEFAULT_POSTGRES_CHANNEL
        self.channel = _validate_identifier(derived_channel[:63], label='channel name')
        self._inbound_bus = EventBus(name=name or f'PostgresEventBridge_{uuid7str()[-8:]}', max_history_size=0)

        self._running = False
        self._write_conn: Any | None = None
        self._listen_conn: Any | None = None
        self._listener_callback: Any | None = None
        self._start_task: asyncio.Task[None] | None = None
        self._start_lock = asyncio.Lock()
        self._listen_query_lock = asyncio.Lock()
        self._table_columns: set[str] = {'event_id', 'event_created_at', 'event_type', _EVENT_PAYLOAD_COLUMN}

    def on(self, event_pattern: EventPatternType, handler: Callable[[BaseEvent[Any]], Any]) -> None:
        self._ensure_started()
        self._inbound_bus.on(event_pattern, handler)

    async def emit(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        self._ensure_started()
        if self._write_conn is None:
            await self.start()

        payload = event.model_dump(mode='json')
        event_fields, event_payload = _split_bridge_payload(payload)
        write_payload = {**event_fields, _EVENT_PAYLOAD_COLUMN: event_payload}
        payload_keys = sorted(write_payload.keys())
        await self._ensure_columns(payload_keys)

        columns_sql = ', '.join(f'"{key}"' for key in payload_keys)
        placeholders_sql = ', '.join(f'${index}' for index in range(1, len(payload_keys) + 1))
        values = [
            json.dumps(write_payload[key], separators=(',', ':')) if write_payload[key] is not None else None
            for key in payload_keys
        ]

        update_fields = [key for key in payload_keys if key != 'event_id']
        if update_fields:
            updates_sql = ', '.join(f'"{key}" = EXCLUDED."{key}"' for key in update_fields)
            upsert_sql = (
                f'INSERT INTO "{self.table}" ({columns_sql}) VALUES ({placeholders_sql}) '
                f'ON CONFLICT ("event_id") DO UPDATE SET {updates_sql}'
            )
        else:
            upsert_sql = (
                f'INSERT INTO "{self.table}" ({columns_sql}) VALUES ({placeholders_sql}) ON CONFLICT ("event_id") DO NOTHING'
            )

        assert self._write_conn is not None
        await self._write_conn.execute(upsert_sql, *values)
        event_id_payload = json.dumps(payload['event_id'], separators=(',', ':'))
        await self._write_conn.execute('SELECT pg_notify($1, $2)', self.channel, event_id_payload)

        if in_handler_context():
            return None
        return event

    async def dispatch(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        return await self.emit(event)

    async def start(self) -> None:
        current_task = asyncio.current_task()
        if self._start_task is not None and self._start_task is not current_task and not self._start_task.done():
            await self._start_task
            return

        if self._running:
            return

        try:
            async with self._start_lock:
                if self._running:
                    return

                asyncpg = self._load_asyncpg()
                write_conn = await asyncpg.connect(self.dsn)
                listen_conn = await asyncpg.connect(self.dsn)
                listener_callback: Any | None = None
                try:
                    self._write_conn = write_conn
                    self._listen_conn = listen_conn
                    await self._ensure_table_exists()
                    await self._refresh_column_cache()
                    await self._ensure_columns(['event_id', 'event_created_at', 'event_type', _EVENT_PAYLOAD_COLUMN])
                    await self._ensure_base_indexes()

                    async def _dispatch_event_id(event_id: str) -> None:
                        try:
                            await self._dispatch_by_event_id(event_id)
                        except Exception:
                            return

                    def _listener(_connection: Any, _pid: int, _channel: str, payload: str) -> None:
                        asyncio.create_task(_dispatch_event_id(payload))

                    listener_callback = _listener
                    await listen_conn.add_listener(self.channel, listener_callback)
                    self._listener_callback = listener_callback
                    self._running = True
                except Exception:
                    if listener_callback is not None:
                        try:
                            await listen_conn.remove_listener(self.channel, listener_callback)
                        except Exception:
                            pass
                    try:
                        await listen_conn.close()
                    except Exception:
                        pass
                    try:
                        await write_conn.close()
                    except Exception:
                        pass
                    if self._listen_conn is listen_conn:
                        self._listen_conn = None
                    if self._write_conn is write_conn:
                        self._write_conn = None
                    if self._listener_callback is listener_callback:
                        self._listener_callback = None
                    raise
        finally:
            if self._start_task is current_task:
                self._start_task = None

    async def close(self, *, clear: bool = True) -> None:
        if self._start_task is not None:
            self._start_task.cancel()
            await asyncio.gather(self._start_task, return_exceptions=True)
            self._start_task = None
        self._running = False
        if self._listen_conn is not None:
            if self._listener_callback is not None:
                try:
                    await self._listen_conn.remove_listener(self.channel, self._listener_callback)
                except Exception:
                    pass
                self._listener_callback = None
            try:
                await self._listen_conn.close()
            except Exception:
                pass
            self._listen_conn = None
        if self._write_conn is not None:
            try:
                await self._write_conn.close()
            except Exception:
                pass
            self._write_conn = None
        await self._inbound_bus.destroy(clear=clear)

    def _ensure_started(self) -> None:
        if self._running:
            return
        if self._start_task is not None and not self._start_task.done():
            return
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return
        self._start_task = asyncio.create_task(self.start())

    async def _dispatch_by_event_id(self, event_id: str) -> None:
        async with self._listen_query_lock:
            assert self._listen_conn is not None
            row = await self._listen_conn.fetchrow(f'SELECT * FROM "{self.table}" WHERE "event_id" = $1', event_id)
            if row is None:
                return

            row_values = dict(row)
            payload: dict[str, Any] = {}

            raw_event_payload = row_values.get(_EVENT_PAYLOAD_COLUMN)
            if isinstance(raw_event_payload, str):
                try:
                    parsed_event_payload: Any = json.loads(raw_event_payload)
                    if _is_str_keyed_dict(parsed_event_payload):
                        payload.update(parsed_event_payload)
                except Exception:
                    pass

            for key, raw_value in row_values.items():
                if key == _EVENT_PAYLOAD_COLUMN or raw_value is None:
                    continue
                if not key.startswith('event_'):
                    continue
                try:
                    decoded_value = json.loads(raw_value)
                except Exception:
                    decoded_value = raw_value

                payload[key] = decoded_value

        await self._dispatch_inbound_payload(payload)

    async def _dispatch_inbound_payload(self, payload: Any) -> None:
        event = BaseEvent[Any].model_validate(payload).event_reset()
        self._inbound_bus.emit(event)

    async def _ensure_table_exists(self) -> None:
        assert self._write_conn is not None
        await self._write_conn.execute(
            f'''
            CREATE TABLE IF NOT EXISTS "{self.table}" (
                "event_id" TEXT PRIMARY KEY,
                "event_created_at" TEXT,
                "event_type" TEXT,
                "event_payload" TEXT
            )
            '''
        )

    async def _ensure_base_indexes(self) -> None:
        assert self._write_conn is not None
        event_created_at_idx = _index_name(self.table, 'event_created_at_idx')
        event_type_idx = _index_name(self.table, 'event_type_idx')

        await self._write_conn.execute(
            f'CREATE INDEX IF NOT EXISTS "{event_created_at_idx}" ON "{self.table}" ("event_created_at")'
        )
        await self._write_conn.execute(f'CREATE INDEX IF NOT EXISTS "{event_type_idx}" ON "{self.table}" ("event_type")')

    async def _refresh_column_cache(self) -> None:
        assert self._write_conn is not None
        rows = await self._write_conn.fetch(
            """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_schema = 'public' AND table_name = $1
            """,
            self.table,
        )
        self._table_columns = {str(row['column_name']) for row in rows}

    async def _ensure_columns(self, keys: list[str]) -> None:
        for key in keys:
            _validate_identifier(key, label='event field name')
            if key != _EVENT_PAYLOAD_COLUMN and not key.startswith('event_'):
                raise ValueError(f'Invalid event field name for bridge column: {key!r}. Only event_* fields become columns')

        missing_columns = [key for key in keys if key not in self._table_columns]
        if not missing_columns:
            return

        assert self._write_conn is not None
        for key in missing_columns:
            await self._write_conn.execute(f'ALTER TABLE "{self.table}" ADD COLUMN IF NOT EXISTS "{key}" TEXT')
            self._table_columns.add(key)

    @staticmethod
    def _load_asyncpg() -> Any:
        try:
            return importlib.import_module('asyncpg')
        except ModuleNotFoundError as exc:
            raise RuntimeError('PostgresEventBridge requires optional dependency: pip install asyncpg') from exc



"""Event bus library."""

from . import events_suck
from .base_event import (
    BaseEvent,
    EventConcurrencyMode,
    EventHandlerCompletionMode,
    EventHandlerConcurrencyMode,
    EventResult,
    EventStatus,
    PythonIdentifierStr,
    PythonIdStr,
    UUIDStr,
)
from .bridges import EventBridge, HTTPEventBridge, JSONLEventBridge, SocketEventBridge, SQLiteEventBridge
from .event_bus import EventBus
from .event_handler import (
    EventHandler,
    EventHandlerAbortedError,
    EventHandlerCancelledError,
    EventHandlerResultSchemaError,
    EventHandlerTimeoutError,
)
from .event_history import EventHistory
from .helpers import monotonic_datetime
from .middlewares import (
    AutoErrorEventMiddleware,
    AutoHandlerChangeEventMiddleware,
    AutoReturnEventMiddleware,
    BusHandlerRegisteredEvent,
    BusHandlerUnregisteredEvent,
    EventBusMiddleware,
    LoggerEventBusMiddleware,
    SQLiteHistoryMirrorMiddleware,
    WALEventBusMiddleware,
)

__all__ = [
    'EventBus',
    'EventBusMiddleware',
    'BusHandlerRegisteredEvent',
    'BusHandlerUnregisteredEvent',
    'EventBridge',
    'HTTPEventBridge',
    'SocketEventBridge',
    'JSONLEventBridge',
    'SQLiteEventBridge',
    'LoggerEventBusMiddleware',
    'SQLiteHistoryMirrorMiddleware',
    'AutoErrorEventMiddleware',
    'AutoHandlerChangeEventMiddleware',
    'AutoReturnEventMiddleware',
    'WALEventBusMiddleware',
    'EventHistory',
    'monotonic_datetime',
    'BaseEvent',
    'EventStatus',
    'EventResult',
    'EventHandler',
    'EventHandlerCancelledError',
    'EventHandlerResultSchemaError',
    'EventHandlerTimeoutError',
    'EventHandlerAbortedError',
    'EventHandlerConcurrencyMode',
    'EventHandlerCompletionMode',
    'EventConcurrencyMode',
    'UUIDStr',
    'PythonIdStr',
    'PythonIdentifierStr',
    'events_suck',
]



"""SQLite flat-table bridge for forwarding events between runtimes.

Uses Python stdlib sqlite3 and polling for new rows.
Schema mirrors Postgres bridge shape:
- event_id (PRIMARY KEY)
- event_created_at (indexed)
- event_type (indexed)
- event_payload (full event JSON payload)
- one TEXT column per event_* field storing JSON-serialized values
"""

from __future__ import annotations

import asyncio
import json
import re
import sqlite3
import time
from collections.abc import Callable
from contextlib import closing
from pathlib import Path
from typing import Any, TypeGuard

from uuid_extensions import uuid7str

from abxbus.base_event import BaseEvent
from abxbus.event_bus import EventBus, EventPatternType, in_handler_context

_IDENTIFIER_RE = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')
_EVENT_PAYLOAD_COLUMN = 'event_payload'


def _validate_identifier(identifier: str, *, label: str) -> str:
    if not _IDENTIFIER_RE.match(identifier):
        raise ValueError(f'Invalid {label}: {identifier!r}. Use only [A-Za-z0-9_] and start with a letter/_')
    return identifier


def _is_str_keyed_dict(value: Any) -> TypeGuard[dict[str, Any]]:
    return isinstance(value, dict)


def _split_bridge_payload(payload: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any]]:
    event_fields: dict[str, Any] = {}
    event_payload: dict[str, Any] = dict(payload)
    for key, value in payload.items():
        if key.startswith('event_'):
            event_fields[key] = value
    return event_fields, event_payload


class SQLiteEventBridge:
    def __init__(
        self,
        path: str,
        table: str = 'abxbus_events',
        *,
        poll_interval: float = 0.25,
        name: str | None = None,
    ):
        self.path = Path(path)
        self.table = _validate_identifier(table, label='table name')
        self.poll_interval = poll_interval
        self._inbound_bus = EventBus(name=name or f'SQLiteEventBridge_{uuid7str()[-8:]}', max_history_size=0)

        self._running = False
        self._start_task: asyncio.Task[None] | None = None
        self._start_lock = asyncio.Lock()
        self._listener_task: asyncio.Task[None] | None = None
        self._last_seen_event_created_at = ''
        self._last_seen_event_id = ''
        self._table_columns: set[str] = {'event_id', 'event_created_at', 'event_type', _EVENT_PAYLOAD_COLUMN}

    def on(self, event_pattern: EventPatternType, handler: Callable[[BaseEvent[Any]], Any]) -> None:
        self._ensure_started()
        self._inbound_bus.on(event_pattern, handler)

    async def emit(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        self._ensure_started()
        if not self._running:
            await self.start()

        payload = event.model_dump(mode='json')
        event_fields, event_payload = _split_bridge_payload(payload)
        write_payload = {**event_fields, _EVENT_PAYLOAD_COLUMN: event_payload}
        payload_keys = sorted(write_payload.keys())

        await asyncio.to_thread(self._ensure_columns, payload_keys)
        await asyncio.to_thread(self._upsert_payload, write_payload, payload_keys)

        if in_handler_context():
            return None
        return event

    async def dispatch(self, event: BaseEvent[Any]) -> BaseEvent[Any] | None:
        return await self.emit(event)

    async def start(self) -> None:
        current_task = asyncio.current_task()
        if self._start_task is not None and self._start_task is not current_task and not self._start_task.done():
            await self._start_task
            return

        if self._running:
            return

        try:
            async with self._start_lock:
                if self._running:
                    return
                self.path.parent.mkdir(parents=True, exist_ok=True)
                await asyncio.to_thread(self._init_db)
                await asyncio.to_thread(self._refresh_column_cache)
                await asyncio.to_thread(
                    self._ensure_columns, ['event_id', 'event_created_at', 'event_type', _EVENT_PAYLOAD_COLUMN]
                )
                await asyncio.to_thread(self._ensure_base_indexes)
                await asyncio.to_thread(self._set_cursor_to_latest_row)
                self._running = True
                if self._listener_task is None or self._listener_task.done():
                    self._listener_task = asyncio.create_task(self._listen_loop())
        finally:
            if self._start_task is current_task:
                self._start_task = None

    async def close(self, *, clear: bool = True) -> None:
        self._running = False
        if self._start_task is not None:
            self._start_task.cancel()
            await asyncio.gather(self._start_task, return_exceptions=True)
            self._start_task = None
        if self._listener_task is not None:
            self._listener_task.cancel()
            await asyncio.gather(self._listener_task, return_exceptions=True)
            self._listener_task = None
        await self._inbound_bus.destroy(clear=clear)

    def _ensure_started(self) -> None:
        if self._running:
            return
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return
        if self._start_task is None or self._start_task.done():
            self._start_task = asyncio.create_task(self.start())

    async def _listen_loop(self) -> None:
        while self._running:
            try:
                rows = await asyncio.to_thread(
                    self._fetch_new_rows,
                    self._last_seen_event_created_at,
                    self._last_seen_event_id,
                )
                for row in rows:
                    event_created_at = str(row.get('event_created_at') or '')
                    event_id = str(row.get('event_id') or '')
                    if event_created_at or event_id:
                        self._last_seen_event_created_at = event_created_at
                        self._last_seen_event_id = event_id

                    payload: dict[str, Any] = {}
                    raw_event_payload = row.get(_EVENT_PAYLOAD_COLUMN)
                    if isinstance(raw_event_payload, str):
                        try:
                            decoded_event_payload: Any = json.loads(raw_event_payload)
                            if _is_str_keyed_dict(decoded_event_payload):
                                payload.update(decoded_event_payload)
                        except Exception:
                            pass

                    for key, raw_value in row.items():
                        if key == _EVENT_PAYLOAD_COLUMN or raw_value is None or not key.startswith('event_'):
                            continue
                        if isinstance(raw_value, str):
                            try:
                                payload[key] = json.loads(raw_value)
                            except Exception:
                                payload[key] = raw_value
                        else:
                            payload[key] = raw_value

                    await self._dispatch_inbound_payload(payload)
            except asyncio.CancelledError:
                raise
            except Exception:
                pass
            await asyncio.sleep(self.poll_interval)

    async def _dispatch_inbound_payload(self, payload: Any) -> None:
        event = BaseEvent[Any].model_validate(payload).event_reset()
        self._inbound_bus.emit(event)

    def _connect(self) -> sqlite3.Connection:
        # Under concurrent bridge startup/teardown across processes, sqlite can
        # intermittently fail with "unable to open database file" while the
        # parent path is being materialized. Recover by ensuring parent exists
        # and retrying a bounded number of times.
        connect_attempts = 20
        conn: sqlite3.Connection | None = None
        last_error: sqlite3.OperationalError | None = None
        for _ in range(connect_attempts):
            try:
                conn = sqlite3.connect(str(self.path), timeout=30.0)
                break
            except sqlite3.OperationalError as exc:
                message = str(exc).lower()
                if 'unable to open database file' not in message:
                    raise
                last_error = exc
                self.path.parent.mkdir(parents=True, exist_ok=True)
                time.sleep(0.05)
        if conn is None:
            assert last_error is not None
            raise last_error

        conn.execute('PRAGMA busy_timeout=30000')
        for _ in range(20):
            try:
                conn.execute('PRAGMA journal_mode=WAL')
                break
            except sqlite3.OperationalError as exc:
                if 'locked' not in str(exc).lower():
                    raise
                time.sleep(0.05)
        conn.row_factory = sqlite3.Row
        return conn

    def _init_db(self) -> None:
        with closing(self._connect()) as conn:
            conn.execute(
                f'''
                CREATE TABLE IF NOT EXISTS "{self.table}" (
                    "event_id" TEXT PRIMARY KEY,
                    "event_created_at" TEXT,
                    "event_type" TEXT,
                    "event_payload" JSON
                )
                '''
            )
            conn.commit()

    def _refresh_column_cache(self) -> None:
        with closing(self._connect()) as conn:
            rows = conn.execute(f'PRAGMA table_info("{self.table}")').fetchall()
            self._table_columns = {str(row['name']) for row in rows}

    def _ensure_columns(self, keys: list[str]) -> None:
        for key in keys:
            _validate_identifier(key, label='event field name')
            if key != _EVENT_PAYLOAD_COLUMN and not key.startswith('event_'):
                raise ValueError(f'Invalid event field name for bridge column: {key!r}. Only event_* fields become columns')

        missing_columns = [key for key in keys if key not in self._table_columns]
        if not missing_columns:
            return

        with closing(self._connect()) as conn:
            for key in missing_columns:
                column_type = 'JSON' if key == _EVENT_PAYLOAD_COLUMN else 'TEXT'
                conn.execute(f'ALTER TABLE "{self.table}" ADD COLUMN "{key}" {column_type}')
                self._table_columns.add(key)
            conn.commit()

    def _ensure_base_indexes(self) -> None:
        event_created_at_index = f'{self.table}_event_created_at_idx'
        event_type_index = f'{self.table}_event_type_idx'

        with closing(self._connect()) as conn:
            conn.execute(f'CREATE INDEX IF NOT EXISTS "{event_created_at_index}" ON "{self.table}" ("event_created_at")')
            conn.execute(f'CREATE INDEX IF NOT EXISTS "{event_type_index}" ON "{self.table}" ("event_type")')
            conn.commit()

    def _upsert_payload(self, payload: dict[str, Any], payload_keys: list[str]) -> None:
        columns_sql = ', '.join(f'"{key}"' for key in payload_keys)
        placeholders_sql = ', '.join('json(?)' if key == _EVENT_PAYLOAD_COLUMN else '?' for key in payload_keys)
        values = [json.dumps(payload[key], separators=(',', ':')) if payload[key] is not None else None for key in payload_keys]

        update_fields = [key for key in payload_keys if key != 'event_id']
        if update_fields:
            updates_sql = ', '.join(f'"{key}" = excluded."{key}"' for key in update_fields)
            upsert_sql = (
                f'INSERT INTO "{self.table}" ({columns_sql}) VALUES ({placeholders_sql}) '
                f'ON CONFLICT("event_id") DO UPDATE SET {updates_sql}'
            )
        else:
            upsert_sql = (
                f'INSERT INTO "{self.table}" ({columns_sql}) VALUES ({placeholders_sql}) ON CONFLICT("event_id") DO NOTHING'
            )

        with closing(self._connect()) as conn:
            conn.execute(upsert_sql, values)
            conn.commit()

    def _set_cursor_to_latest_row(self) -> None:
        with closing(self._connect()) as conn:
            row = conn.execute(
                f'''
                SELECT
                    COALESCE("event_created_at", '') AS event_created_at,
                    COALESCE("event_id", '') AS event_id
                FROM "{self.table}"
                ORDER BY COALESCE("event_created_at", '') DESC, COALESCE("event_id", '') DESC
                LIMIT 1
                '''
            ).fetchone()
            if row is None:
                self._last_seen_event_created_at = ''
                self._last_seen_event_id = ''
                return
            self._last_seen_event_created_at = str(row['event_created_at'] or '')
            self._last_seen_event_id = str(row['event_id'] or '')

    def _fetch_new_rows(self, last_event_created_at: str, last_event_id: str) -> list[dict[str, Any]]:
        with closing(self._connect()) as conn:
            rows = conn.execute(
                f'''
                SELECT *
                FROM "{self.table}"
                WHERE
                    COALESCE("event_created_at", '') > ?
                    OR (
                        COALESCE("event_created_at", '') = ?
                        AND COALESCE("event_id", '') > ?
                    )
                ORDER BY COALESCE("event_created_at", '') ASC, COALESCE("event_id", '') ASC
                ''',
                (last_event_created_at, last_event_created_at, last_event_id),
            ).fetchall()
            return [dict(row) for row in rows]



import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence
from typing import Any, TypeAlias, cast

from pydantic import BaseModel, Field, TypeAdapter, create_model

_SCHEMA_TYPE_REGISTRY: tuple[tuple[str, type[Any], str], ...] = (
    ('string', str, 'string'),
    ('integer', int, 'number'),  # note both integer and number are mapped to the same JSON Schema type
    ('number', float, 'number'),
    ('boolean', bool, 'boolean'),
    ('object', dict, 'object'),
    ('array', list, 'array'),
    ('null', type(None), 'null'),
)

TYPE_MAPPING: dict[str, type[Any]] = {schema_type: python_type for schema_type, python_type, _ in _SCHEMA_TYPE_REGISTRY}

CONSTRAINT_MAPPING: dict[str, str] = {
    'minimum': 'ge',
    'maximum': 'le',
    'exclusiveMinimum': 'gt',
    'exclusiveMaximum': 'lt',
    'inclusiveMinimum': 'ge',
    'inclusiveMaximum': 'le',
    'minItems': 'min_length',
    'maxItems': 'max_length',
}

_NON_PRIMITIVE_SCHEMA_TYPES = {'object', 'array'}

PRIMITIVE_TYPE_MAPPING: dict[str, type[Any]] = {
    schema_type: python_type
    for schema_type, python_type, _ in _SCHEMA_TYPE_REGISTRY
    if schema_type not in _NON_PRIMITIVE_SCHEMA_TYPES
}

IDENTIFIER_NORMALIZATION: dict[str, str] = {schema_type: identifier for schema_type, _, identifier in _SCHEMA_TYPE_REGISTRY}

JSON_SCHEMA_DRAFT = 'https://json-schema.org/draft/2020-12/schema'
_TYPE_ADAPTER_CACHE: dict[Any, TypeAdapter[Any]] = {}

FieldDefinition: TypeAlias = Any | tuple[Any, Any]


def _get_cached_type_adapter(result_type: Any) -> TypeAdapter[Any]:
    """Return a cached TypeAdapter for hashable result types."""
    try:
        cached = _TYPE_ADAPTER_CACHE.get(result_type)
    except TypeError:
        return TypeAdapter(result_type)
    if cached is not None:
        return cached
    adapter = TypeAdapter(result_type)
    _TYPE_ADAPTER_CACHE[result_type] = adapter
    return adapter


def _as_string_key_dict(value: object) -> dict[str, Any] | None:
    """Return a dict view with only string keys, otherwise None."""
    if not isinstance(value, Mapping):
        return None
    value_mapping = cast(Mapping[object, Any], value)
    normalized: dict[str, Any] = {}
    for raw_key, raw_value in value_mapping.items():
        if isinstance(raw_key, str):
            normalized[raw_key] = raw_value
    return normalized


def _as_non_string_sequence(value: object) -> Sequence[Any] | None:
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        return cast(Sequence[Any], value)
    return None


def _iter_string_key_dicts(value: object) -> Iterator[dict[str, Any]]:
    sequence_values = _as_non_string_sequence(value)
    if sequence_values is None:
        return
    for candidate_raw in sequence_values:
        candidate = _as_string_key_dict(candidate_raw)
        if candidate is not None:
            yield candidate


def _extract_non_null_json_schema_type(schema: Mapping[str, Any]) -> str | None:
    raw_type = schema.get('type')
    if isinstance(raw_type, str):
        return raw_type

    raw_type_values = _as_non_string_sequence(raw_type)
    if raw_type_values is not None:
        non_null_types = [item for item in raw_type_values if isinstance(item, str) and item != 'null']
        if len(non_null_types) == 1:
            return non_null_types[0]

    return None


def _json_schema_allows_null(schema: Mapping[str, Any]) -> bool:
    raw_type = schema.get('type')
    if raw_type == 'null':
        return True
    raw_type_values = _as_non_string_sequence(raw_type)
    if raw_type_values is not None:
        if any(item == 'null' for item in raw_type_values):
            return True

    for candidate in _iter_string_key_dicts(schema.get('anyOf')):
        if candidate.get('type') == 'null':
            return True
    return False


def _nullable_type(resolved_type: Any, *, nullable: bool) -> Any:
    if not nullable or resolved_type is type(None):
        return resolved_type
    return resolved_type | None


def normalize_result_dict(value: Any) -> dict[str, Any]:
    """Return a dict with only string keys from an arbitrary mapping-like value."""
    return _as_string_key_dict(value) or {}


def _json_schema_primitive_type(schema: dict[str, Any]) -> type[Any] | None:
    """Map simple JSON Schema primitive types to Python runtime types."""
    schema_type = _extract_non_null_json_schema_type(schema)
    return PRIMITIVE_TYPE_MAPPING.get(schema_type) if schema_type is not None else None


def _json_schema_identifier(schema: dict[str, Any]) -> str | None:
    schema_type = _extract_non_null_json_schema_type(schema)
    return IDENTIFIER_NORMALIZATION.get(schema_type) if schema_type is not None else None


def get_field_params_from_field_schema(field_schema: dict[str, Any]) -> dict[str, Any]:
    """Gets Pydantic field parameters from a JSON schema field."""
    field_params: dict[str, Any] = {}
    for constraint, constraint_value in CONSTRAINT_MAPPING.items():
        if constraint in field_schema:
            field_params[constraint_value] = field_schema[constraint]
    if 'description' in field_schema:
        field_params['description'] = field_schema['description']
    if 'default' in field_schema:
        field_params['default'] = field_schema['default']
    return field_params


def _json_schema_ref_name(schema: Mapping[str, Any]) -> str | None:
    raw_ref = schema.get('$ref')
    if raw_ref is None:
        return None
    reference = str(raw_ref).strip()
    if not reference:
        return None
    return reference.split('/')[-1]


def _build_model_fields_from_schema(
    schema: Mapping[str, Any],
    *,
    resolve_field_type: Callable[[dict[str, Any]], Any],
) -> dict[str, FieldDefinition]:
    fields: dict[str, FieldDefinition] = {}
    properties = _as_string_key_dict(schema.get('properties'))
    if properties is None:
        return fields
    required_raw = schema.get('required')
    required_fields: set[str] = set()
    required_values = _as_non_string_sequence(required_raw)
    if required_values is not None:
        required_fields = {name for name in required_values if isinstance(name, str)}

    for field_name, field_schema_raw in properties.items():
        field_schema = _as_string_key_dict(field_schema_raw)
        if field_schema is None:
            continue
        field_type = resolve_field_type(field_schema)
        field_params = get_field_params_from_field_schema(field_schema=field_schema)
        field_name_str = str(field_name)
        is_required = field_name_str in required_fields
        has_default = 'default' in field_params
        if not is_required and not has_default:
            relaxed_type = _nullable_type(field_type, nullable=True)
            fields[field_name_str] = (relaxed_type, Field(default=None, **field_params))
        else:
            fields[field_name_str] = (field_type, Field(**field_params))

    return fields


def _create_dynamic_model(
    *,
    model_name: str,
    model_schema: Mapping[str, Any],
    fields: Mapping[str, FieldDefinition],
) -> type[BaseModel]:
    field_definitions: dict[str, Any] = dict(fields)
    return create_model(
        model_name,
        __doc__=str(model_schema.get('description', '')),
        **field_definitions,
    )


def pydantic_model_from_json_schema(result_type: Any) -> Any:
    """Reconstruct runtime types from JSON Schema when possible."""
    if not isinstance(result_type, dict):
        return result_type
    normalized_schema = normalize_result_dict(result_type)
    definitions = _as_string_key_dict(normalized_schema.get('$defs')) or {}
    models: dict[str, type[BaseModel]] = {}
    model_build_stack: set[str] = set()

    def _combine_union_types(resolved_types: list[Any], *, nullable: bool) -> Any:
        if not resolved_types:
            return _nullable_type(Any, nullable=nullable)
        combined = resolved_types[0]
        for candidate_type in resolved_types[1:]:
            combined = combined | candidate_type
        return _nullable_type(combined, nullable=nullable)

    def _resolve_ref_model(model_reference: str) -> Any:
        if model_reference in models:
            return models[model_reference]
        if model_reference in model_build_stack:
            return Any
        model_schema_raw = definitions.get(model_reference)
        model_schema = _as_string_key_dict(model_schema_raw)
        if model_schema is None:
            return Any

        model_build_stack.add(model_reference)
        try:
            dynamic_model = _create_dynamic_model(
                model_name=model_reference,
                model_schema=model_schema,
                fields=_build_model_fields_from_schema(
                    model_schema,
                    resolve_field_type=_resolve_schema,
                ),
            )
            models[model_reference] = dynamic_model
            return dynamic_model
        finally:
            model_build_stack.remove(model_reference)

    def _resolve_array_schema(schema: dict[str, Any], *, nullable: bool) -> Any:
        prefix_items_raw = schema.get('prefixItems')
        prefix_items = _as_non_string_sequence(prefix_items_raw)
        if prefix_items is not None:
            tuple_items = [_resolve_schema(item) for item in prefix_items]
            if tuple_items:
                resolved_tuple = tuple.__class_getitem__(tuple(tuple_items))
                return _nullable_type(resolved_tuple, nullable=nullable)

        items_schema = _as_string_key_dict(schema.get('items'))
        if items_schema is None:
            return _nullable_type(list[Any], nullable=nullable)
        item_type = _resolve_schema(items_schema)
        if schema.get('uniqueItems') is True:
            return _nullable_type(set[item_type], nullable=nullable)
        return _nullable_type(list[item_type], nullable=nullable)

    def _resolve_object_schema(schema: dict[str, Any], *, nullable: bool) -> Any:
        properties = _as_string_key_dict(schema.get('properties'))
        if properties:
            dynamic_model = _create_dynamic_model(
                model_name=str(schema.get('title', 'InlineObject')),
                model_schema=schema,
                fields=_build_model_fields_from_schema(
                    schema,
                    resolve_field_type=_resolve_schema,
                ),
            )
            return _nullable_type(dynamic_model, nullable=nullable)

        additional_properties = _as_string_key_dict(schema.get('additionalProperties'))
        if additional_properties is not None:
            value_type = _resolve_schema(additional_properties)
            return _nullable_type(dict[str, value_type], nullable=nullable)
        return _nullable_type(dict[str, Any], nullable=nullable)

    def _resolve_schema(schema_raw: Any) -> Any:
        schema = normalize_result_dict(schema_raw)
        if not schema:
            return Any

        allows_null = _json_schema_allows_null(schema)
        model_reference = _json_schema_ref_name(schema)
        if model_reference is not None:
            return _nullable_type(_resolve_ref_model(model_reference), nullable=allows_null)

        primitive_type = _json_schema_primitive_type(schema)
        if primitive_type is not None:
            return _nullable_type(primitive_type, nullable=allows_null)

        any_of_candidates = _as_non_string_sequence(schema.get('anyOf'))
        if any_of_candidates is not None:
            resolved_types: list[Any] = []
            includes_null = allows_null
            for candidate in _iter_string_key_dicts(any_of_candidates):
                if candidate.get('type') == 'null':
                    includes_null = True
                    continue
                resolved_types.append(_resolve_schema(candidate))
            return _combine_union_types(resolved_types, nullable=includes_null)

        schema_type = _extract_non_null_json_schema_type(schema)
        if schema_type == 'null':
            return type(None)
        if schema_type == 'array':
            return _resolve_array_schema(schema, nullable=allows_null)
        if schema_type == 'object':
            return _resolve_object_schema(schema, nullable=allows_null)
        if isinstance(schema_type, str) and schema_type in TYPE_MAPPING:
            return _nullable_type(TYPE_MAPPING[schema_type], nullable=allows_null)
        return _nullable_type(Any, nullable=allows_null)

    for model_name in definitions:
        _resolve_ref_model(model_name)
    return _resolve_schema(normalized_schema)


def pydantic_model_to_json_schema(result_type: Any) -> dict[str, Any] | None:
    """Best-effort conversion of a Python result schema/type into JSON Schema."""
    if result_type is None:
        return None
    if isinstance(result_type, dict):
        schema = dict(cast(dict[str, Any], result_type))
        schema.setdefault('$schema', JSON_SCHEMA_DRAFT)
        return schema
    if isinstance(result_type, str):
        return None

    try:
        if inspect.isclass(result_type) and issubclass(result_type, BaseModel):
            schema = result_type.model_json_schema()
            schema.setdefault('$schema', JSON_SCHEMA_DRAFT)
            return schema
    except TypeError:
        pass

    try:
        schema = TypeAdapter(result_type).json_schema()
        normalized_schema = normalize_result_dict(schema)
        normalized_schema.setdefault('$schema', JSON_SCHEMA_DRAFT)
        return normalized_schema
    except Exception:
        return None


def result_type_identifier_from_schema(result_type: Any) -> str | None:
    if result_type is None:
        return None
    if isinstance(result_type, str):
        return result_type
    if isinstance(result_type, dict):
        return _json_schema_identifier(normalize_result_dict(result_type))

    if result_type is str:
        return 'string'
    if result_type in (int, float):
        return 'number'
    if result_type is bool:
        return 'boolean'

    derived_schema = pydantic_model_to_json_schema(result_type)
    if isinstance(derived_schema, dict):
        return _json_schema_identifier(derived_schema)
    return None


def validate_result_against_type(result_type: Any, result: Any) -> Any:
    if result_type is None:
        return result

    if isinstance(result_type, dict):
        result_type = pydantic_model_from_json_schema(result_type)

    if inspect.isclass(result_type) and issubclass(result_type, BaseModel):
        return result_type.model_validate(result)

    adapter = _get_cached_type_adapter(result_type)
    return adapter.validate_python(result)


__all__ = [
    'get_field_params_from_field_schema',
    'normalize_result_dict',
    'pydantic_model_from_json_schema',
    'pydantic_model_to_json_schema',
    'result_type_identifier_from_schema',
    'validate_result_against_type',
]



from __future__ import annotations

import inspect
import types
from collections.abc import Awaitable, Callable, Mapping
from types import SimpleNamespace
from typing import Any, Protocol, TypeVar, cast, get_args, get_origin

from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from abxbus.base_event import BaseEvent
from abxbus.event_bus import EventBus
from abxbus.helpers import extract_basemodel_generic_arg

EventClass = type[BaseEvent[Any]]
_BASE_EVENT_FIELD_NAMES = frozenset(BaseEvent.model_fields)
_EMPTY = inspect.Parameter.empty
T_Result = TypeVar('T_Result')


class _HasBus(Protocol):
    bus: EventBus


class GeneratedEvents(SimpleNamespace):
    by_name: dict[str, EventClass]


def _custom_event_fields(event_cls: EventClass) -> list[tuple[str, FieldInfo]]:
    return [
        (field_name, field) for field_name, field in event_cls.model_fields.items() if field_name not in _BASE_EVENT_FIELD_NAMES
    ]


def _event_field_default(field: FieldInfo) -> Any:
    default = field.get_default(call_default_factory=False)
    if default is PydanticUndefined:
        return None
    return default


def _event_result_annotation(event_cls: EventClass) -> Any:
    generic_result_type = extract_basemodel_generic_arg(event_cls)
    if generic_result_type is not None:
        return generic_result_type

    result_field = event_cls.model_fields.get('event_result_type')
    if result_field is not None and result_field.default not in (None, PydanticUndefined):
        return result_field.default

    return Any


def _callable_params(func: Callable[..., Any]) -> tuple[list[inspect.Parameter], bool, Any]:
    signature = inspect.signature(func)
    params = list(signature.parameters.values())
    has_var_kwargs = any(param.kind == inspect.Parameter.VAR_KEYWORD for param in params)
    if params and params[0].name in {'self', 'cls'}:
        params = params[1:]

    filtered: list[inspect.Parameter] = []
    for param in params:
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            raise TypeError(f'events_suck does not support *args in {func!r}')
        if param.kind == inspect.Parameter.POSITIONAL_ONLY:
            raise TypeError(f'events_suck does not support positional-only params in {func!r}')
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        filtered.append(param)

    return_annotation = signature.return_annotation if signature.return_annotation is not _EMPTY else Any
    return filtered, has_var_kwargs, return_annotation


def _event_payload(event: BaseEvent[Any]) -> dict[str, Any]:
    payload = {
        field_name: getattr(event, field_name)
        for field_name in event.__class__.model_fields
        if field_name not in _BASE_EVENT_FIELD_NAMES
    }
    extras = event.model_extra
    if isinstance(extras, dict):
        payload.update(extras)
    return payload


def _annotation_allows_none(annotation: Any) -> bool:
    if annotation is None or annotation is type(None):
        return True
    origin = get_origin(annotation)
    if origin is None:
        return False
    return any(arg is type(None) for arg in get_args(annotation))


def _make_event_class(event_name: str, func: Callable[..., Any]) -> EventClass:
    if not event_name.isidentifier() or event_name.startswith('_'):
        raise ValueError(f'Invalid event name: {event_name!r}')

    params, _, return_annotation = _callable_params(func)
    annotations: dict[str, Any] = {'event_result_type': Any}
    namespace: dict[str, Any] = {'__module__': __name__, 'event_type': event_name, 'event_result_type': return_annotation}
    for param in params:
        annotation = param.annotation if param.annotation is not _EMPTY else Any
        annotations[param.name] = annotation
        if param.default is not _EMPTY:
            namespace[param.name] = param.default
        elif _annotation_allows_none(annotation):
            namespace[param.name] = None
    namespace['__annotations__'] = annotations
    try:
        event_base = cast(type[Any], BaseEvent[return_annotation])
    except Exception:
        event_base = BaseEvent
    event_cls = types.new_class(event_name, (event_base,), exec_body=lambda ns: ns.update(namespace))
    return cast(EventClass, event_cls)


def make_events(events: Mapping[str, Callable[..., Any]]) -> GeneratedEvents:
    by_name = {event_name: _make_event_class(event_name, func) for event_name, func in events.items()}
    return GeneratedEvents(**by_name, by_name=by_name)


def make_handler(func: Callable[..., object]) -> Callable[[BaseEvent[Any]], Awaitable[object]]:
    params, has_var_kwargs, _ = _callable_params(func)

    async def _handler(event: BaseEvent[Any]) -> object:
        payload = _event_payload(event)
        kwargs: dict[str, Any] = {}
        for param in params:
            if param.name in payload:
                kwargs[param.name] = payload.pop(param.name)
            elif param.default is _EMPTY:
                raise TypeError(f'Missing required event field {param.name!r} for handler {func!r}')
        if has_var_kwargs:
            kwargs.update(payload)
        result = func(**kwargs)
        if inspect.isawaitable(result):
            return await result
        return result

    return _handler


def _build_event_method(class_name: str, method_name: str, event_cls: EventClass):
    event_fields = _custom_event_fields(event_cls)
    event_field_names = tuple(field_name for field_name, _ in event_fields)

    parameters = [inspect.Parameter('self', inspect.Parameter.POSITIONAL_OR_KEYWORD)]
    for field_name, field in event_fields:
        field_annotation = field.annotation if field.annotation is not None else Any
        field_default = inspect.Parameter.empty if field.is_required() else _event_field_default(field)
        parameters.append(
            inspect.Parameter(
                field_name,
                inspect.Parameter.KEYWORD_ONLY,
                default=field_default,
                annotation=field_annotation,
            )
        )
    parameters.append(inspect.Parameter('extra', inspect.Parameter.VAR_KEYWORD, annotation=Any))
    signature = inspect.Signature(parameters=parameters, return_annotation=_event_result_annotation(event_cls))

    async def _method(self: _HasBus, *args: Any, **kwargs: Any) -> Any:
        bound = signature.bind(self, *args, **kwargs)
        payload: dict[str, Any] = {
            field_name: bound.arguments[field_name] for field_name in event_field_names if field_name in bound.arguments
        }
        payload.update(cast(dict[str, Any], bound.arguments.get('extra', {})))
        return await self.bus.emit(event_cls(**payload)).event_result()

    _method.__name__ = method_name
    _method.__qualname__ = f'{class_name}.{method_name}'
    _method.__annotations__ = {
        **{field_name: (field.annotation if field.annotation is not None else Any) for field_name, field in event_fields},
        'extra': Any,
        'return': signature.return_annotation,
    }
    cast(Any, _method).__signature__ = signature
    return _method


def wrap(class_name: str, methods: Mapping[str, EventClass]) -> type[Any]:
    if not class_name.isidentifier() or class_name.startswith('_'):
        raise ValueError(f'Invalid class name: {class_name!r}')

    def __init__(self: _HasBus, bus: EventBus | None = None) -> None:
        self.bus = bus or EventBus(f'{class_name}Bus')

    namespace: dict[str, Any] = {
        '__module__': __name__,
        '__annotations__': {'bus': EventBus},
        '__init__': __init__,
    }

    for method_name, event_cls in methods.items():
        if not method_name.isidentifier() or method_name.startswith('_'):
            raise ValueError(f'Invalid method name: {method_name!r}')
        namespace[method_name] = _build_event_method(class_name, method_name, event_cls)

    return cast(type[Any], type(class_name, (), namespace))


__all__ = ['GeneratedEvents', 'make_events', 'make_handler', 'wrap']



import asyncio
import contextvars
import inspect
import json
import logging
import warnings
import weakref
from collections import defaultdict
from collections.abc import Callable, Coroutine, Iterator, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta
from functools import partial
from typing import Any, Literal, TypeVar, overload
from uuid import UUID

from uuid_extensions import uuid7str

uuid7str: Callable[[], str] = uuid7str

from abxbus.base_event import (
    ABXBUS_LOGGING_LEVEL,
    BaseEvent,
    EventConcurrencyMode,
    EventHandlerCompletionMode,
    EventHandlerConcurrencyMode,
    EventResult,
    EventResultsDict,
    EventStatus,
    PythonIdentifierStr,
    PythonIdStr,
    T_Event,
    T_EventResultType,
    UUIDStr,
)
from abxbus.event_handler import (
    ContravariantEventHandlerCallable,
    EventHandler,
    EventHandlerAbortedError,
    EventHandlerCallable,
    EventHandlerCancelledError,
    EventHandlerTimeoutError,
)
from abxbus.event_history import EventHistory
from abxbus.helpers import (
    CleanShutdownQueue,
    QueueShutDown,
    _run_with_slow_monitor,  # pyright: ignore[reportPrivateUsage]
    log_filtered_traceback,
)
from abxbus.lock_manager import LockManager, LockManagerProtocol, ReentrantLock
from abxbus.middlewares import EventBusMiddleware

logger = logging.getLogger('abxbus')
logger.setLevel(ABXBUS_LOGGING_LEVEL)

T_ExpectedEvent = TypeVar('T_ExpectedEvent', bound=BaseEvent[Any])
T_OnEvent = TypeVar('T_OnEvent', bound=BaseEvent[Any])

EventPatternType = PythonIdentifierStr | Literal['*'] | type[BaseEvent[Any]]
# Middleware entries can be provided as already-initialized instances or middleware classes.
EventBusMiddlewareInput = EventBusMiddleware | type[EventBusMiddleware]


def _as_any(value: Any) -> Any:
    return value


@dataclass(slots=True, eq=False)
class _FindWaiter:
    event_key: str
    matches: Callable[[BaseEvent[Any]], bool]
    future: asyncio.Future[BaseEvent[Any] | None]
    timeout_handle: asyncio.TimerHandle | None = None


class GlobalEventBusRegistry:
    """Weak global registry of EventBus instances."""

    def __init__(self) -> None:
        self._buses: weakref.WeakSet['EventBus'] = weakref.WeakSet()

    def add(self, bus: 'EventBus') -> None:
        self._buses.add(bus)

    def discard(self, bus: 'EventBus') -> None:
        self._buses.discard(bus)

    def has(self, bus: 'EventBus') -> bool:
        return bus in self._buses

    @property
    def size(self) -> int:
        return len(self._buses)

    def __iter__(self) -> Iterator['EventBus']:
        return iter(self._buses)

    def __len__(self) -> int:
        return len(self._buses)

    def __contains__(self, bus: object) -> bool:
        return bus in self._buses


def get_current_event() -> BaseEvent[Any] | None:
    """Return the currently active event in this async context, if any."""
    return EventBus.current_event_context.get()


def get_current_handler_id() -> str | None:
    """Return the currently active handler id in this async context, if any."""
    return EventBus.current_handler_id_context.get()


def get_current_eventbus() -> 'EventBus | None':
    """Return the currently active EventBus in this async context, if any."""
    return EventBus.current_eventbus_context.get()


def in_handler_context() -> bool:
    """Return True when called from inside an executing handler context."""
    return get_current_handler_id() is not None


def current_handler_context_is_stale() -> bool:
    """Return True when this async context belongs to a cancelled/aborted handler."""
    current_event = get_current_event()
    current_handler_id = get_current_handler_id()
    if current_event is None or current_handler_id is None:
        return False
    current_task = asyncio.current_task()
    if current_task is not None and current_task.cancelling():
        return True
    current_result = current_event.event_results.get(current_handler_id)
    if current_result is None:
        return False
    current_error = current_result.error
    if current_error is None and isinstance(current_result.result, BaseException):
        current_error = current_result.result
    return isinstance(
        current_error,
        (EventHandlerTimeoutError, EventHandlerCancelledError, EventHandlerAbortedError),
    )


class EventBus:
    """
    Async event bus with write-ahead logging and guaranteed FIFO processing.

    Features:
    - Enqueue events synchronously, await their results using 'await Event()'
    - FIFP Write-ahead logging with UUIDs and timestamps,
    - Serial event processing, configurable handler concurrency per event ('serial' | 'parallel')
    """

    # Track all EventBus instances (using weakrefs to allow garbage collection)
    all_instances: GlobalEventBusRegistry = GlobalEventBusRegistry()
    _lock_for_event_global_serial: ReentrantLock = ReentrantLock()

    # Per-loop EventBus registry and original close method tracking.
    _loop_eventbus_instances: weakref.WeakKeyDictionary[asyncio.AbstractEventLoop, weakref.WeakSet['EventBus']] = (
        weakref.WeakKeyDictionary()
    )
    _loop_original_close: weakref.WeakKeyDictionary[asyncio.AbstractEventLoop, Callable[[], None]] = weakref.WeakKeyDictionary()

    # Context variables for current event/handler/bus execution scope.
    current_event_context: ContextVar[BaseEvent[Any] | None] = ContextVar('current_event', default=None)
    current_handler_id_context: ContextVar[str | None] = ContextVar('current_handler_id', default=None)
    current_eventbus_context: ContextVar['EventBus | None'] = ContextVar('current_eventbus', default=None)

    # Class Attributes
    name: PythonIdentifierStr = 'EventBus'
    event_concurrency: EventConcurrencyMode = EventConcurrencyMode.BUS_SERIAL
    event_timeout: float | None = 60.0
    event_slow_timeout: float | None = 300.0
    event_handler_concurrency: EventHandlerConcurrencyMode = EventHandlerConcurrencyMode.SERIAL
    event_handler_completion: EventHandlerCompletionMode = EventHandlerCompletionMode.ALL
    event_handler_slow_timeout: float | None = 30.0
    event_handler_detect_file_paths: bool = True
    warn_on_duplicate_handler_names: bool = True
    max_handler_recursion_depth: int = 2

    # Runtime State
    id: UUIDStr = '00000000-0000-0000-0000-000000000000'
    handlers: dict[PythonIdStr, EventHandler]
    handlers_by_key: dict[str, list[PythonIdStr]]
    pending_event_queue: CleanShutdownQueue[BaseEvent[Any]] | None
    event_history: EventHistory[BaseEvent[Any]]

    _is_running: bool = False
    _runloop_task: asyncio.Task[None] | None = None
    _parallel_event_tasks: set[asyncio.Task[None]]
    _pending_middleware_tasks: set[asyncio.Task[None]]
    _on_idle: asyncio.Event | None = None
    in_flight_event_ids: set[str]
    processing_event_ids: set[str]
    _duplicate_handler_name_check_limit: int = 256
    _pending_handler_changes: list[tuple[EventHandler, bool]]
    find_waiters: set[_FindWaiter]
    _lock_for_event_bus_serial: ReentrantLock
    locks: LockManagerProtocol
    _destroyed: bool

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # Subclasses get isolated global registries/locks by default.
        cls.all_instances = GlobalEventBusRegistry()
        cls._lock_for_event_global_serial = ReentrantLock()
        cls._loop_eventbus_instances = weakref.WeakKeyDictionary()
        cls._loop_original_close = weakref.WeakKeyDictionary()

    @classmethod
    def iter_all_instances(cls) -> Iterator['EventBus']:
        """Iterate live EventBus instances for ``cls`` and all subclasses."""
        pending_classes: list[type[EventBus]] = [cls]
        seen_classes: set[type[EventBus]] = set()
        seen_buses: set[int] = set()
        while pending_classes:
            bus_class = pending_classes.pop()
            if bus_class in seen_classes:
                continue
            seen_classes.add(bus_class)
            pending_classes.extend(bus_class.__subclasses__())
            for bus in list(bus_class.all_instances):
                bus_id = id(bus)
                if bus_id in seen_buses:
                    continue
                seen_buses.add(bus_id)
                yield bus

    def __init__(
        self,
        name: PythonIdentifierStr | None = None,
        event_concurrency: EventConcurrencyMode | str | None = None,
        event_handler_concurrency: EventHandlerConcurrencyMode | str = EventHandlerConcurrencyMode.SERIAL,
        event_handler_completion: EventHandlerCompletionMode | str = EventHandlerCompletionMode.ALL,
        max_history_size: int | None = 100,  # Keep only 100 events in history
        max_history_drop: bool = False,
        event_timeout: float | None = 60.0,
        event_slow_timeout: float | None = 300.0,
        event_handler_slow_timeout: float | None = 30.0,
        event_handler_detect_file_paths: bool = True,
        warn_on_duplicate_handler_names: bool = True,
        max_handler_recursion_depth: int = 2,
        middlewares: Sequence[EventBusMiddlewareInput] | None = None,
        id: UUIDStr | str | None = None,
    ):
        self.id = str(UUID(str(id))) if id is not None else uuid7str()
        self.name = name or f'{self.__class__.__name__}_{self.id[-8:]}'
        assert self.name.isidentifier(), f'EventBus name must be a unique identifier string, got: {self.name}'

        # Force garbage collection to clean up any dead EventBus instances in the WeakSet
        # gc.collect()  # Commented out - this is expensive and causes 5s delays when creating many EventBus instances

        # Check for name uniqueness among existing instances
        # We'll collect potential conflicts and check if they're still alive
        original_name = self.name
        conflicting_buses: list[EventBus] = []

        for existing_bus in list(type(self).all_instances):  # Make a list copy to avoid modification during iteration
            if existing_bus is not self and existing_bus.name == self.name:
                # Since destroy() renames buses to _destroyed_{id}, any bus with a matching
                # user-specified name is either running or never-started - both should
                # be considered conflicts. This makes name conflict detection deterministic.
                conflicting_buses.append(existing_bus)

        # If we found conflicting buses, auto-generate a unique suffix
        if conflicting_buses:
            # Generate a unique suffix using the last 8 chars of a UUID
            unique_suffix = uuid7str()[-8:]
            self.name = f'{original_name}_{unique_suffix}'

            warnings.warn(
                f'⚠️ EventBus with name "{original_name}" already exists. '
                f'Auto-generated unique name: "{self.name}" to avoid conflicts. '
                f'Consider using unique names or destroy(clear=True) on unused buses.',
                UserWarning,
                stacklevel=2,
            )

        self.pending_event_queue = None
        self.event_history = EventHistory(max_history_size=max_history_size, max_history_drop=max_history_drop)
        self.handlers = {}
        self.handlers_by_key = defaultdict(list)
        self._lock_for_event_bus_serial = ReentrantLock()
        self.locks = LockManager()
        self._parallel_event_tasks = set()
        self._pending_middleware_tasks = set()
        try:
            self.event_concurrency = EventConcurrencyMode(event_concurrency or EventConcurrencyMode.BUS_SERIAL)
        except ValueError as exc:
            raise AssertionError(
                f'event_concurrency must be "global-serial", "bus-serial", or "parallel", got: {event_concurrency!r}'
            ) from exc
        try:
            self.event_handler_concurrency = EventHandlerConcurrencyMode(
                event_handler_concurrency or EventHandlerConcurrencyMode.SERIAL
            )
        except ValueError as exc:
            raise AssertionError(
                f'event_handler_concurrency must be "serial" or "parallel", got: {event_handler_concurrency!r}'
            ) from exc
        try:
            self.event_handler_completion = EventHandlerCompletionMode(event_handler_completion or EventHandlerCompletionMode.ALL)
        except ValueError as exc:
            raise AssertionError(f'event_handler_completion must be "all" or "first", got: {event_handler_completion!r}') from exc
        self.event_timeout = 60.0 if event_timeout is None else event_timeout
        self.event_slow_timeout = 300.0 if event_slow_timeout is None else event_slow_timeout
        self.event_handler_slow_timeout = 30.0 if event_handler_slow_timeout is None else event_handler_slow_timeout
        self.event_handler_detect_file_paths = bool(event_handler_detect_file_paths)
        self.warn_on_duplicate_handler_names = bool(warn_on_duplicate_handler_names)
        self.max_handler_recursion_depth = int(max_handler_recursion_depth)
        assert self.event_timeout is None or self.event_timeout >= 0, (
            f'event_timeout must be >= 0 or None, got: {self.event_timeout!r}'
        )
        assert self.event_slow_timeout is None or self.event_slow_timeout >= 0, (
            f'event_slow_timeout must be >= 0 or None, got: {self.event_slow_timeout!r}'
        )
        assert self.event_handler_slow_timeout is None or self.event_handler_slow_timeout >= 0, (
            f'event_handler_slow_timeout must be >= 0 or None, got: {self.event_handler_slow_timeout!r}'
        )
        assert self.max_handler_recursion_depth >= 0, (
            f'max_handler_recursion_depth must be >= 0, got: {self.max_handler_recursion_depth!r}'
        )
        self._on_idle = None
        self.middlewares = self._normalize_middlewares(middlewares)
        self.in_flight_event_ids = set()
        self.processing_event_ids = set()
        self._pending_handler_changes = []
        self.find_waiters = set()
        self._destroyed = False

        # Register this instance
        type(self).all_instances.add(self)

    def _raise_if_destroyed(self) -> None:
        if self._destroyed:
            raise RuntimeError(f'{self} has been destroyed and cannot be used again')

    @staticmethod
    def _normalize_middlewares(middlewares: Sequence[EventBusMiddlewareInput] | None) -> list[EventBusMiddleware]:
        """Normalize middleware inputs to concrete middleware instances.

        Accepts mixed instance/class entries and instantiates class entries once
        during bus construction, preserving registration order.
        """
        normalized: list[EventBusMiddleware] = []
        for middleware in middlewares or ():
            if isinstance(middleware, EventBusMiddleware):
                normalized.append(middleware)
                continue
            normalized.append(middleware())
        return normalized

    def __del__(self):
        """Auto-cleanup on garbage collection"""
        # Most cleanup should have been done by the event loop close hook
        # This is just a fallback for any remaining cleanup

        # Signal the run loop to exit
        self._is_running = False
        if self.pending_event_queue:
            try:
                # Wake any blocked queue.get() in the weak runloop so it can exit.
                self.pending_event_queue.shutdown(immediate=True)
            except Exception:
                pass
        if self._runloop_task and not self._runloop_task.done():
            try:
                setattr(self._runloop_task, '_log_destroy_pending', False)
                self._runloop_task.cancel()
            except Exception:
                pass
        for task in tuple(self._parallel_event_tasks):
            if task.done():
                continue
            try:
                task.cancel()
            except Exception:
                pass
        for task in tuple(self._pending_middleware_tasks):
            if task.done():
                continue
            try:
                task.cancel()
            except Exception:
                pass

        # Our custom queue handles cleanup properly in shutdown()
        # No need for manual cleanup here

        # Check total memory usage across all EventBus instances
        try:
            self._check_total_memory_usage()
        except Exception:
            # Don't let memory check errors prevent cleanup
            pass

    def __str__(self) -> str:
        icon = '🟢' if self._is_running else '🔴'
        queue_size = self.pending_event_queue.qsize() if self.pending_event_queue else 0
        return f'{self.label}{icon}(queue={queue_size} active={len(self.in_flight_event_ids)} history={len(self.event_history)} handlers={len(self.handlers)})'

    @property
    def label(self) -> str:
        return f'{self.name}#{self.id[-4:]}'

    def __repr__(self) -> str:
        return str(self)

    @property
    def event_bus_serial_lock(self) -> ReentrantLock:
        """Public accessor for the bus-serial event lock used by LockManager."""
        return self._lock_for_event_bus_serial

    @property
    def event_global_serial_lock(self) -> ReentrantLock:
        """Class-scoped global-serial lock (isolated per EventBus subclass)."""
        return type(self)._lock_for_event_global_serial

    @staticmethod
    def _stub_handler_callable(_event: BaseEvent[Any]) -> None:
        return None

    @classmethod
    def _event_json_payload(cls, event: BaseEvent[Any]) -> dict[str, Any]:
        return event.model_dump(mode='json')

    @staticmethod
    def _normalize_json_object(value: Any) -> dict[str, Any] | None:
        if not isinstance(value, dict):
            return None
        value_any = _as_any(value)
        normalized: dict[str, Any] = {}
        for raw_key, raw_value in value_any.items():
            key_value: Any = raw_key
            item_value: Any = raw_value
            if isinstance(key_value, str):
                normalized[key_value] = item_value
        return normalized

    @classmethod
    def _normalize_json_object_required(cls, value: Any, error_message: str) -> dict[str, Any]:
        normalized = cls._normalize_json_object(value)
        if normalized is None:
            raise TypeError(error_message)
        return normalized

    @staticmethod
    def _normalize_json_string_list(value: Any) -> list[str] | None:
        if not isinstance(value, list):
            return None
        value_any = _as_any(value)
        normalized: list[str] = []
        for raw_item in value_any:
            item_value: Any = raw_item
            if isinstance(item_value, str):
                normalized.append(item_value)
        return normalized

    @staticmethod
    def _upsert_handler_index(handlers_by_key: dict[str, list[PythonIdStr]], event_pattern: str, handler_id: PythonIdStr) -> None:
        ids = handlers_by_key.setdefault(event_pattern, [])
        if handler_id not in ids:
            ids.append(handler_id)

    @classmethod
    def _hydrate_handler_from_event_result(
        cls, bus: 'EventBus', event: BaseEvent[Any], result_payload: dict[str, Any]
    ) -> EventHandler:
        if isinstance(result_payload.get('handler'), dict):
            raise ValueError('Legacy nested EventResult.handler payload is not supported')
        raw_handler_id = result_payload.get('handler_id')
        handler_id: str | None = raw_handler_id if isinstance(raw_handler_id, str) and raw_handler_id else None
        if handler_id is None:
            raise ValueError('EventResult JSON payload must include handler_id')
        if handler_id in bus.handlers:
            return bus.handlers[handler_id]
        handler_payload: dict[str, Any] = {'id': handler_id}
        handler_payload.setdefault('handler_name', result_payload.get('handler_name') or 'anonymous')
        handler_payload.setdefault('handler_file_path', result_payload.get('handler_file_path'))
        if result_payload.get('handler_timeout') is not None:
            handler_payload.setdefault('handler_timeout', result_payload.get('handler_timeout'))
        if result_payload.get('handler_slow_timeout') is not None:
            handler_payload.setdefault('handler_slow_timeout', result_payload.get('handler_slow_timeout'))
        if result_payload.get('handler_registered_at') is not None:
            handler_payload.setdefault('handler_registered_at', result_payload.get('handler_registered_at'))
        handler_payload.setdefault('event_pattern', result_payload.get('handler_event_pattern') or event.event_type)
        handler_payload.setdefault('eventbus_name', result_payload.get('eventbus_name') or bus.name)
        handler_payload.setdefault('eventbus_id', result_payload.get('eventbus_id') or bus.id)

        handler_entry = EventHandler.model_validate({**handler_payload, 'handler': cls._stub_handler_callable})
        resolved_handler_id = handler_entry.id
        bus.handlers[resolved_handler_id] = handler_entry
        cls._upsert_handler_index(bus.handlers_by_key, handler_entry.event_pattern, resolved_handler_id)
        return handler_entry

    @classmethod
    def _hydrate_event_result_from_json(
        cls,
        *,
        bus: 'EventBus',
        event: BaseEvent[Any],
        result_payload: dict[str, Any],
    ) -> tuple[EventResult[Any], list[str]]:
        handler_entry = cls._hydrate_handler_from_event_result(bus, event, result_payload)
        model_payload = dict(result_payload)
        child_ids = [child_id for child_id in model_payload.pop('event_children', []) if isinstance(child_id, str)]
        has_result = 'result' in model_payload
        raw_result = model_payload.pop('result', None)
        for flat_key in (
            'handler_id',
            'handler_name',
            'handler_file_path',
            'handler_timeout',
            'handler_slow_timeout',
            'handler_registered_at',
            'handler_event_pattern',
            'eventbus_name',
            'eventbus_id',
        ):
            model_payload.pop(flat_key, None)
        model_payload['event_id'] = event.event_id
        model_payload['handler'] = handler_entry
        event_result = EventResult.model_validate(model_payload)
        if has_result:
            event_result.result = raw_result
        event_result.handler = handler_entry
        return event_result, child_ids

    def model_dump(self) -> dict[str, Any]:
        handlers_payload: dict[str, dict[str, Any]] = {}
        for handler_entry in self.handlers.values():
            handlers_payload[handler_entry.id] = handler_entry.model_dump(mode='json', exclude={'handler'})

        handlers_by_key_payload: dict[str, list[str]] = {
            str(key): [str(handler_id) for handler_id in handler_ids] for key, handler_ids in self.handlers_by_key.items()
        }
        for handler_entry in self.handlers.values():
            self._upsert_handler_index(handlers_by_key_payload, handler_entry.event_pattern, handler_entry.id)

        event_history_payload: dict[str, dict[str, Any]] = {}
        for event in self.event_history.values():
            event_history_payload[event.event_id] = self._event_json_payload(event)

        pending_event_ids: list[str] = []
        if self.pending_event_queue is not None:
            for queued_event in self.pending_event_queue.iter_items():
                event_id = queued_event.event_id
                if event_id not in event_history_payload:
                    event_history_payload[event_id] = self._event_json_payload(queued_event)
                pending_event_ids.append(event_id)

        return {
            'id': self.id,
            'name': self.name,
            'max_history_size': self.event_history.max_history_size,
            'max_history_drop': self.event_history.max_history_drop,
            'event_concurrency': str(self.event_concurrency),
            'event_timeout': self.event_timeout,
            'event_slow_timeout': self.event_slow_timeout,
            'event_handler_concurrency': str(self.event_handler_concurrency),
            'event_handler_completion': str(self.event_handler_completion),
            'event_handler_slow_timeout': self.event_handler_slow_timeout,
            'event_handler_detect_file_paths': self.event_handler_detect_file_paths,
            'handlers': handlers_payload,
            'handlers_by_key': handlers_by_key_payload,
            'event_history': event_history_payload,
            'pending_event_queue': pending_event_ids,
        }

    def model_dump_json(self, *, indent: int | None = None) -> str:
        return json.dumps(self.model_dump(), ensure_ascii=False, default=str, indent=indent)

    @classmethod
    def validate(cls, data: Any) -> 'EventBus':
        raw_payload: Any = data
        if isinstance(data, (bytes, bytearray)):
            raw_payload = data.decode('utf-8')
        if isinstance(raw_payload, str):
            raw_payload = json.loads(raw_payload)
        payload = cls._normalize_json_object(raw_payload)
        if payload is None:
            raise TypeError(f'EventBus.validate() expects dict or JSON string, got: {type(data).__name__}')
        name = payload.get('name')
        requested_name = str(name) if isinstance(name, str) else None
        bus = cls(
            name=None,
            event_concurrency=payload.get('event_concurrency'),
            event_handler_concurrency=payload.get('event_handler_concurrency') or EventHandlerConcurrencyMode.SERIAL,
            event_handler_completion=payload.get('event_handler_completion') or EventHandlerCompletionMode.ALL,
            max_history_size=payload.get('max_history_size')
            if isinstance(payload.get('max_history_size'), int) or payload.get('max_history_size') is None
            else 100,
            max_history_drop=bool(payload.get('max_history_drop', False)),
            event_timeout=payload.get('event_timeout')
            if isinstance(payload.get('event_timeout'), (int, float)) or payload.get('event_timeout') is None
            else None,
            event_slow_timeout=payload.get('event_slow_timeout')
            if isinstance(payload.get('event_slow_timeout'), (int, float)) or payload.get('event_slow_timeout') is None
            else 300.0,
            event_handler_slow_timeout=payload.get('event_handler_slow_timeout')
            if isinstance(payload.get('event_handler_slow_timeout'), (int, float))
            or payload.get('event_handler_slow_timeout') is None
            else 30.0,
            event_handler_detect_file_paths=bool(payload.get('event_handler_detect_file_paths', True)),
            warn_on_duplicate_handler_names=bool(payload.get('warn_on_duplicate_handler_names', True)),
            id=payload.get('id') if isinstance(payload.get('id'), str) else None,
        )
        if requested_name is not None:
            bus.name = requested_name

        bus.handlers.clear()
        bus.handlers_by_key = defaultdict(list)
        bus.event_history.clear()

        raw_handlers = cls._normalize_json_object_required(
            payload.get('handlers', {}),
            'EventBus.validate() expects handlers to be an id-keyed object',
        )
        for raw_handler_id, raw_handler_payload in raw_handlers.items():
            handler_payload_json = cls._normalize_json_object(raw_handler_payload)
            if handler_payload_json is None:
                continue
            handler_payload: dict[str, Any] = dict(handler_payload_json)
            if 'id' not in handler_payload:
                handler_payload['id'] = raw_handler_id
            handler_entry = EventHandler.model_validate({**handler_payload, 'handler': cls._stub_handler_callable})
            bus.handlers[handler_entry.id] = handler_entry

        raw_handlers_by_key = cls._normalize_json_object_required(
            payload.get('handlers_by_key', {}),
            'EventBus.validate() expects handlers_by_key to be an object',
        )
        for raw_key, raw_ids in raw_handlers_by_key.items():
            handler_id_list = cls._normalize_json_string_list(raw_ids)
            if handler_id_list is None:
                continue
            handler_ids: list[PythonIdStr] = []
            for handler_id in handler_id_list:
                handler_ids.append(handler_id)
            bus.handlers_by_key[raw_key] = handler_ids

        pending_child_links: list[tuple[EventResult[Any], list[str]]] = []

        raw_event_history = cls._normalize_json_object_required(
            payload.get('event_history'),
            'EventBus.validate() expects event_history to be an id-keyed object',
        )
        history_items: list[tuple[str, Any]] = list(raw_event_history.items())

        for event_id_hint, event_payload_any in history_items:
            event_payload_json = cls._normalize_json_object(event_payload_any)
            if event_payload_json is None:
                continue
            event_payload: dict[str, Any] = dict(event_payload_json)
            raw_event_results = event_payload.pop('event_results', {})
            if not isinstance(raw_event_results, dict):
                raise TypeError('EventBus.validate() expects event_results to be an id-keyed object')
            if 'event_id' not in event_payload or not isinstance(event_payload.get('event_id'), str):
                event_payload['event_id'] = event_id_hint
            try:
                event = BaseEvent[Any].model_validate(event_payload)
            except Exception:
                continue

            hydrated_results: dict[PythonIdStr, EventResult[Any]] = {}
            result_items: list[dict[str, Any]] = []
            raw_event_results_any = _as_any(raw_event_results)
            for raw_handler_id, raw_item in raw_event_results_any.items():
                result_payload = cls._normalize_json_object(raw_item)
                if result_payload is None:
                    continue
                item = dict(result_payload)
                item.setdefault('handler_id', str(raw_handler_id))
                result_items.append(item)

            for result_payload in result_items:
                try:
                    event_result, child_ids = cls._hydrate_event_result_from_json(
                        bus=bus,
                        event=event,
                        result_payload=result_payload,
                    )
                except Exception:
                    continue
                hydrated_results[event_result.handler_id] = event_result
                pending_child_links.append((event_result, child_ids))

            event.event_results = EventResultsDict(hydrated_results)
            bus.event_history[event.event_id] = event

        pending_event_ids: list[str] = []
        raw_pending_queue = cls._normalize_json_string_list(payload.get('pending_event_queue'))
        if raw_pending_queue is None:
            raise TypeError('EventBus.validate() expects pending_event_queue to be a list of event ids')
        pending_event_ids.extend(raw_pending_queue)

        for event_result, child_ids in pending_child_links:
            event_result.event_children = [bus.event_history[child_id] for child_id in child_ids if child_id in bus.event_history]

        if pending_event_ids:
            queue = CleanShutdownQueue[BaseEvent[Any]](maxsize=0)
            for event_id in pending_event_ids:
                event = bus.event_history.get(event_id)
                if event is None:
                    continue
                queue.put_nowait(event)
            bus.pending_event_queue = queue
            bus._on_idle = asyncio.Event()
            if queue.qsize() == 0 and not bus._has_inflight_events_fast():
                bus._on_idle.set()
            else:
                bus._on_idle.clear()
        else:
            bus.pending_event_queue = None
            bus._on_idle = None

        bus._is_running = False
        bus._runloop_task = None
        bus._parallel_event_tasks = set()
        bus._pending_middleware_tasks = set()
        bus.in_flight_event_ids = set()
        bus.processing_event_ids = set()
        return bus

    async def on_event_change(self, event: BaseEvent[Any], status: EventStatus) -> None:
        if not self.middlewares:
            return
        for middleware in self.middlewares:
            await middleware.on_event_change(self, event, status)

    async def on_event_result_change(self, event: BaseEvent[Any], event_result: EventResult[Any], status: EventStatus) -> None:
        if not self.middlewares:
            return
        for middleware in self.middlewares:
            await middleware.on_event_result_change(self, event, event_result, status)

    def remove_event_from_pending_queue(self, event: BaseEvent[Any]) -> bool:
        if self.pending_event_queue is None:
            return False
        return self.pending_event_queue.remove_item(event)

    def mark_pending_queue_task_done(self) -> None:
        if self.pending_event_queue is None:
            return
        try:
            self.pending_event_queue.task_done()
        except ValueError:
            pass

    def _start_parallel_event_task_from_queue(self, event: BaseEvent[Any]) -> None:
        """Process a queued parallel event even when the main runloop is inside a handler."""

        async def process_parallel_event() -> None:
            removed = self.remove_event_from_pending_queue(event)
            if not removed:
                return
            try:
                await self.step(event=event)
            finally:
                self.mark_pending_queue_task_done()

        task = asyncio.create_task(
            process_parallel_event(),
            name=f'{self}._process_parallel_event({event.event_id[-4:]})',
            context=contextvars.Context(),
        )
        self._parallel_event_tasks.add(task)

        def _on_done(done_task: asyncio.Task[None]) -> None:
            self._parallel_event_tasks.discard(done_task)
            if self._on_idle and self.pending_event_queue:
                if not self._has_inflight_events_fast() and self.pending_event_queue.qsize() == 0:
                    self._on_idle.set()
            if done_task.cancelled():
                return
            try:
                exc = done_task.exception()
            except asyncio.CancelledError:
                return
            if exc is not None:
                logger.exception('❌ Parallel event task error: %s %s', type(exc).__name__, exc)

        task.add_done_callback(_on_done)

    def queue_contains_event_id(self, event_id: str) -> bool:
        if self.pending_event_queue is None:
            return False
        for queued_event in self.pending_event_queue.iter_items():
            if queued_event.event_id == event_id:
                return True
        return False

    def is_event_inflight_or_queued(self, event_id: str) -> bool:
        if event_id in self.in_flight_event_ids:
            return True
        if event_id in self.processing_event_ids:
            return True
        return self.queue_contains_event_id(event_id)

    def is_event_processing(self, event_id: str) -> bool:
        return event_id in self.processing_event_ids

    def _resolve_find_waiters(self, event: BaseEvent[Any]) -> None:
        if not self.find_waiters:
            return
        for waiter in tuple(self.find_waiters):
            if (waiter.event_key != '*' and event.event_type != waiter.event_key) or not waiter.matches(event):
                continue
            if waiter.timeout_handle is not None:
                waiter.timeout_handle.cancel()
            self.find_waiters.discard(waiter)
            if not waiter.future.done():
                waiter.future.set_result(event)

    async def _wait_for_future_match(
        self,
        event_key: str,
        matches: Callable[[BaseEvent[Any]], bool],
        future: bool | float,
    ) -> BaseEvent[Any] | None:
        if future is False:
            return None
        event_match_future: asyncio.Future[BaseEvent[Any] | None] = asyncio.get_running_loop().create_future()
        waiter = _FindWaiter(event_key=event_key, matches=matches, future=event_match_future)
        if future is not True:
            timeout_seconds = float(future)

            def _on_wait_timeout() -> None:
                self.find_waiters.discard(waiter)
                if waiter.timeout_handle is not None:
                    waiter.timeout_handle.cancel()
                    waiter.timeout_handle = None
                if not event_match_future.done():
                    event_match_future.set_result(None)

            waiter.timeout_handle = asyncio.get_running_loop().call_later(timeout_seconds, _on_wait_timeout)
        self.find_waiters.add(waiter)
        try:
            return await event_match_future
        finally:
            self.find_waiters.discard(waiter)
            if waiter.timeout_handle is not None:
                waiter.timeout_handle.cancel()
                waiter.timeout_handle = None

    async def on_bus_handlers_change(self, handler: EventHandler, registered: bool) -> None:
        if not self.middlewares:
            return
        for middleware in self.middlewares:
            await middleware.on_bus_handlers_change(self, handler, registered)

    def _schedule_middleware_task(self, task: asyncio.Task[None]) -> None:
        self._pending_middleware_tasks.add(task)
        if self._on_idle is not None:
            self._on_idle.clear()

        def _on_done(done_task: asyncio.Task[None]) -> None:
            self._pending_middleware_tasks.discard(done_task)
            if self._on_idle and self.pending_event_queue:
                if not self._has_inflight_events_fast() and self.pending_event_queue.qsize() == 0:
                    self._on_idle.set()
            if done_task.cancelled():
                return
            exc = done_task.exception()
            if exc is not None:
                logger.error('❌ %s middleware task failed: %s(%r)', self, type(exc).__name__, exc)

        task.add_done_callback(_on_done)

    def _notify_handler_change(self, handler: EventHandler, registered: bool) -> None:
        if not self.middlewares:
            return
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # Preserve .on()/.off() notifications registered before an event loop starts.
            self._pending_handler_changes.append((handler.model_copy(deep=False), registered))
            return
        task = loop.create_task(self.on_bus_handlers_change(handler, registered))
        self._schedule_middleware_task(task)

    def _flush_pending_handler_changes(self) -> None:
        if not self._pending_handler_changes or not self.middlewares:
            return
        loop = asyncio.get_running_loop()
        queued = list(self._pending_handler_changes)
        self._pending_handler_changes.clear()
        for handler, registered in queued:
            task = loop.create_task(self.on_bus_handlers_change(handler, registered))
            self._schedule_middleware_task(task)

    @staticmethod
    def _resolve_event_slow_timeout(event: BaseEvent[Any], eventbus: 'EventBus') -> float | None:
        event_slow_timeout = event.event_slow_timeout if event.event_slow_timeout is not None else eventbus.event_slow_timeout
        return event_slow_timeout if event_slow_timeout is not None and event_slow_timeout > 0 else None

    @staticmethod
    def _resolve_handler_slow_timeout(event: BaseEvent[Any], handler: EventHandler, eventbus: 'EventBus') -> float | None:
        if 'handler_slow_timeout' in handler.model_fields_set and handler.handler_slow_timeout is not None:
            return handler.handler_slow_timeout
        event_handler_slow_timeout = (
            event.event_handler_slow_timeout
            if event.event_handler_slow_timeout is not None
            else eventbus.event_handler_slow_timeout
        )
        return event_handler_slow_timeout if event_handler_slow_timeout is not None and event_handler_slow_timeout > 0 else None

    @staticmethod
    def _resolve_handler_timeout(
        event: BaseEvent[Any],
        handler: EventHandler,
        eventbus: 'EventBus',
        timeout_override: float | None = None,
    ) -> float | None:
        if 'handler_timeout' in handler.model_fields_set and handler.handler_timeout is not None:
            resolved_handler_timeout = handler.handler_timeout
        elif event.event_handler_timeout is not None:
            resolved_handler_timeout = event.event_handler_timeout
        else:
            resolved_handler_timeout = None

        if resolved_handler_timeout is not None and resolved_handler_timeout <= 0:
            resolved_handler_timeout = None

        raw_event_timeout = event.event_timeout if event.event_timeout is not None else eventbus.event_timeout
        resolved_event_timeout = raw_event_timeout if raw_event_timeout is not None and raw_event_timeout > 0 else None

        if resolved_handler_timeout is None and resolved_event_timeout is None:
            resolved_timeout = None
        elif resolved_handler_timeout is None:
            resolved_timeout = resolved_event_timeout
        elif resolved_event_timeout is None:
            resolved_timeout = resolved_handler_timeout
        else:
            resolved_timeout = min(resolved_handler_timeout, resolved_event_timeout)

        if timeout_override is not None and timeout_override <= 0:
            timeout_override = None

        if timeout_override is None:
            return resolved_timeout
        if resolved_timeout is None:
            return timeout_override
        return min(resolved_timeout, timeout_override)

    async def _slow_event_warning_monitor(self, event: BaseEvent[Any], event_slow_timeout: float) -> None:
        await asyncio.sleep(event_slow_timeout)
        if event.event_status == EventStatus.COMPLETED:
            return
        running_handler_count = sum(1 for result in event.event_results.values() if result.status == 'started')
        started_at = event.event_started_at or event.event_created_at
        started_at_dt = datetime.fromisoformat(started_at)
        elapsed_seconds = max(0.0, (datetime.now(UTC) - started_at_dt).total_seconds())
        logger.warning(
            '⚠️ Slow event processing: %s.on(%s#%s, %d handlers) still running after %.2fs',
            self.label,
            event.event_type,
            event.event_id[-4:],
            running_handler_count,
            elapsed_seconds,
        )

    def _has_inflight_events_fast(self) -> bool:
        return bool(
            self.in_flight_event_ids or self.processing_event_ids or self._parallel_event_tasks or self._pending_middleware_tasks
        )

    def _mark_event_complete_on_all_buses(self, event: BaseEvent[Any]) -> None:
        event_id = event.event_id
        for bus in list(type(self).all_instances):
            if bus:
                bus.in_flight_event_ids.discard(event_id)
                if bus.event_history.max_history_size == 0:
                    # max_history_size=0 means "keep only in-flight events".
                    # As soon as an event is completed, drop it from history.
                    bus.event_history.remove_event(event_id)

    @property
    def events_pending(self) -> list[BaseEvent[Any]]:
        """Get events that haven't started processing yet (does not include events still being enqueued in self.event_queue)."""
        return [
            event
            for event in self.event_history.values()
            if event.event_status != EventStatus.COMPLETED and event.event_status != EventStatus.STARTED
        ]

    @property
    def events_started(self) -> list[BaseEvent[Any]]:
        """Get events currently being processed"""
        return [
            event
            for event in self.event_history.values()
            if event.event_status != EventStatus.COMPLETED and event.event_status == EventStatus.STARTED
        ]

    @property
    def events_completed(self) -> list[BaseEvent[Any]]:
        """Get events that have completed processing"""
        return [event for event in self.event_history.values() if event.event_status == EventStatus.COMPLETED]

    # Overloads for typed event patterns with specific handler signatures
    # Order matters - more specific types must come before general ones

    # Class pattern registration keeps strict event typing.
    @overload
    def on(self, event_pattern: type[T_OnEvent], handler: ContravariantEventHandlerCallable[T_OnEvent]) -> EventHandler: ...

    # String and wildcard registration is looser: any BaseEvent subclass handler is allowed.
    @overload
    def on(
        self,
        event_pattern: PythonIdentifierStr | Literal['*'],
        handler: ContravariantEventHandlerCallable[T_OnEvent],
    ) -> EventHandler: ...

    # I dont think this is needed, but leaving it here for now
    # 9. Coroutine[Any, Any, Any] - direct coroutine
    # @overload
    # def on(self, event_pattern: EventPatternType, handler: Coroutine[Any, Any, Any]) -> None: ...

    def on(
        self,
        event_pattern: EventPatternType,
        handler: ContravariantEventHandlerCallable[T_OnEvent],
    ) -> EventHandler:
        """
        Subscribe to events matching a pattern, event type name, or event model class.
        Use event_pattern='*' to subscribe to all events. Handler can be sync or async function or method.

        Examples:
                eventbus.on('TaskStartedEvent', handler)  # Specific event type
                eventbus.on(TaskStartedEvent, handler)  # Event model class
                eventbus.on('*', handler)  # Subscribe to all events
                eventbus.on('*', other_eventbus.emit)  # Forward all events to another EventBus

        Note: When forwarding events between buses, all handler results are automatically
        flattened into the original event's results, so EventResults sees all handlers
        from all buses as a single flat collection.
        """
        self._raise_if_destroyed()
        assert isinstance(event_pattern, str) or isinstance(event_pattern, type), (
            f'Invalid event pattern: {event_pattern}, must be a string event type or subclass of BaseEvent'
        )
        assert inspect.isfunction(handler) or inspect.ismethod(handler) or inspect.iscoroutinefunction(handler), (
            f'Invalid handler: {handler}, must be a sync or async function or method'
        )

        # Normalize event key to string event_type or wildcard.
        event_key = EventHistory.normalize_event_pattern(event_pattern)

        # Ensure event_key is definitely a string at this point
        assert isinstance(event_key, str)

        # Check for duplicate handler names. Keep this bounded so large handler
        # registrations (e.g. perf scenarios with tens of thousands of handlers)
        # do not degrade into O(n^2) registration time.
        new_handler_name = EventHandler.get_callable_handler_name(handler)
        existing_handler_ids = self.handlers_by_key.get(event_key, [])
        if (
            self.warn_on_duplicate_handler_names
            and existing_handler_ids
            and len(existing_handler_ids) <= self._duplicate_handler_name_check_limit
        ):
            for existing_handler_id in existing_handler_ids:
                existing_handler = self.handlers.get(existing_handler_id)
                if existing_handler and existing_handler.handler_name == new_handler_name:
                    warnings.warn(
                        f"⚠️ {self} Handler {new_handler_name} already registered for event '{event_key}'. "
                        f'This may make it difficult to filter event results by handler name. '
                        f'Consider using unique function names.',
                        UserWarning,
                        stacklevel=2,
                    )
                    break

        # Register handler entry and index it by event key.
        handler_entry = EventHandler.from_callable(
            handler=handler,
            event_pattern=event_key,
            eventbus_name=self.name,
            eventbus_id=self.id,
            detect_handler_file_path=self.event_handler_detect_file_paths,
        )
        assert handler_entry.id is not None
        self.handlers[handler_entry.id] = handler_entry
        self.handlers_by_key[event_key].append(handler_entry.id)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                '👂 %s.on(%s, %s) Registered event handler #%s',
                self,
                event_key,
                handler_entry.handler_name,
                handler_entry.id[-4:],
            )
        self._notify_handler_change(handler_entry, registered=True)
        return handler_entry

    @overload
    def off(
        self, event_pattern: type[T_Event], handler: EventHandlerCallable | PythonIdStr | EventHandler | None = None
    ) -> None: ...

    @overload
    def off(
        self,
        event_pattern: PythonIdentifierStr | Literal['*'],
        handler: EventHandlerCallable | PythonIdStr | EventHandler | None = None,
    ) -> None: ...

    def off(
        self,
        event_pattern: EventPatternType,
        handler: EventHandlerCallable | PythonIdStr | EventHandler | None = None,
    ) -> None:
        """Deregister handlers for an event pattern by id, callable, EventHandler, or all."""
        self._raise_if_destroyed()
        event_key = EventHistory.normalize_event_pattern(event_pattern)
        indexed_ids = list(self.handlers_by_key.get(event_key, []))
        if not indexed_ids:
            return

        requested_id: str | None = None
        requested_callable: EventHandlerCallable | None = None
        if isinstance(handler, EventHandler):
            requested_id = handler.id
        elif isinstance(handler, str):
            requested_id = handler
        elif handler is not None:
            requested_callable = handler

        for handler_id in indexed_ids:
            entry = self.handlers.get(handler_id)
            if entry is None:
                self._remove_indexed_handler(event_key, handler_id)
                continue

            should_remove = False
            if handler is None:
                should_remove = True
            elif requested_id is not None and entry.id == requested_id:
                should_remove = True
            elif requested_callable is not None and entry.handler is requested_callable:
                should_remove = True

            if should_remove:
                self.handlers.pop(handler_id, None)
                self._remove_indexed_handler(event_key, handler_id)
                self._notify_handler_change(entry, registered=False)

    def emit(self, event: T_ExpectedEvent) -> T_ExpectedEvent:
        """
        Enqueue an event for processing and immediately return an Event(status='pending') version (synchronous).
        You can await the returned Event(status='pending') object to block until it is done being executed aka Event(status='completed'),
        or you can interact with the unawaited Event(status='pending') before its handlers have finished.

        (The first EventBus.emit() call auto-starts the bus processing loop task if not already running)

        >>> completed_event = await eventbus.emit(SomeEvent())
                # 1. enqueues the event synchronously
                # 2. returns an awaitable SomeEvent() with pending results in .event_results
                # 3. awaits the SomeEvent() which waits until all pending results are complete and returns the completed SomeEvent()

        >>> result_value = await eventbus.emit(SomeEvent()).event_result()
                # 1. enqueues the event synchronously
                # 2. returns a pending SomeEvent() with pending results in .event_results
                # 3. awaiting .event_result() waits until all pending results are complete, and returns the raw result value of the first one
        """
        self._raise_if_destroyed()

        try:
            asyncio.get_running_loop()
        except RuntimeError:
            raise RuntimeError(f'{self}.emit() called but no event loop is running! Event not queued: {event.event_type}')

        assert event.event_id, 'Missing event.event_id: UUIDStr = uuid7str()'
        assert event.event_created_at, 'Missing event.event_created_at: str = monotonic_datetime()'
        assert event.event_type and event.event_type.isidentifier(), 'Missing event.event_type: str'

        if current_handler_context_is_stale():
            return event

        # Capture emit-time context for propagation to handlers (GitHub issue #20)
        # This ensures ContextVars set before emit() are accessible in handlers
        if event._get_dispatch_context() is None:  # pyright: ignore[reportPrivateUsage]
            event._set_dispatch_context(contextvars.copy_context())  # pyright: ignore[reportPrivateUsage]

        # Add this EventBus label to the event_path if not already there
        if self.label not in event.event_path:
            # preserve identity of the original object instead of creating a new one, so that the original object remains awaitable to get the result
            # NOT: event = event.model_copy(update={'event_path': event.event_path + [self.name]})
            event.event_path.append(self.label)
        else:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    '⚠️ %s.emit(%s) - Bus already in path, not adding again. Path: %s',
                    self,
                    event.event_type,
                    event.event_path,
                )

        assert event.event_path, 'Missing event.event_path: list[str] (with at least one bus label recorded in it)'
        assert all(
            '#' in entry
            and entry.rsplit('#', 1)[0].isidentifier()
            and entry.rsplit('#', 1)[1].isalnum()
            and len(entry.rsplit('#', 1)[1]) == 4
            for entry in event.event_path
        ), f'Event.event_path must be a list of EventBus labels BusName#abcd, got: {event.event_path}'

        # NOTE:
        # emit() is intentionally synchronous and runs on the same event-loop
        # thread as the runloop task. Blocking here for "pressure" would deadlock
        # naive flood loops because the runloop cannot progress until emit() returns.
        # So pressure is handled by policy:
        #   - max_history_drop=True  -> absorb and trim oldest history entries
        #   - max_history_drop=False -> reject new emits at max_history_size
        if (
            self.event_history.max_history_size is not None
            and self.event_history.max_history_size > 0
            and not self.event_history.max_history_drop
        ):
            if len(self.event_history) >= self.event_history.max_history_size:
                # Before rejecting, opportunistically evict already-completed history entries.
                # This preserves max_history_drop=False semantics (never dropping in-flight events)
                # while avoiding needless backpressure when only completed entries are occupying the cap.
                self.event_history.trim_event_history(owner_label=str(self))
            if len(self.event_history) >= self.event_history.max_history_size:
                raise RuntimeError(
                    f'{self} history limit reached ({len(self.event_history)}/{self.event_history.max_history_size}); '
                    'set event_history.max_history_drop=True to drop old history instead of rejecting new events'
                )

        # Auto-start if needed
        self._flush_pending_handler_changes()
        self._start()
        # Ensure every emitted event has a completion signal tied to this loop.
        # Completion logic always sets this signal; consumers like event_results_* await it.
        _ = event.event_completed_signal

        # Put event in queue synchronously using put_nowait
        if self.pending_event_queue:
            try:
                self.pending_event_queue.put_nowait(event)
                # Only add to history after successfully queuing
                self.event_history[event.event_id] = event
                self.in_flight_event_ids.add(event.event_id)
                # Resolve future find waiters immediately on emit so callers
                # don't wait for queue position or handler execution.
                self._resolve_find_waiters(event)
                if self.locks.get_lock_for_event(self, event) is None:
                    self._start_parallel_event_task_from_queue(event)
                if logger.isEnabledFor(logging.INFO):
                    logger.info(
                        '🗣️ %s.emit(%s) ➡️ %s#%s (#%d %s)',
                        self,
                        event.event_type,
                        event.event_type,
                        event.event_id[-4:],
                        self.pending_event_queue.qsize(),
                        event.event_status,
                    )
            except asyncio.QueueFull:
                # Don't add to history if we can't queue it
                logger.error(
                    f'⚠️ {self} Event queue is full! Dropping event and aborting {event.event_type}:\n{event.model_dump_json()}'
                )
                raise  # could also block indefinitely until queue has space, but dont drop silently or delete events
        else:
            logger.warning('⚠️ %s.emit() called but event_queue is None! Event not queued: %s', self, event.event_type)

        # Note: We do NOT pre-create EventResults here anymore.
        # EventResults are created only when handlers actually start executing.
        # This avoids "orphaned" pending results for handlers that get filtered out later.

        # Amortize cleanup work by trimming only after a soft overage; this keeps
        # hot emit fast under large naive floods while still bounding memory.
        if (
            self.event_history.max_history_size is not None
            and self.event_history.max_history_size > 0
            and self.event_history.max_history_drop
        ):
            soft_limit = max(self.event_history.max_history_size, int(self.event_history.max_history_size * 1.2))
            if len(self.event_history) > soft_limit:
                self.event_history.trim_event_history(owner_label=str(self))

        return event

    def dispatch(self, event: T_ExpectedEvent) -> T_ExpectedEvent:
        """Convenience synonym for :meth:`emit`."""
        return self.emit(event)

    def _remove_indexed_handler(self, event_pattern: str, handler_id: PythonIdStr) -> None:
        ids = self.handlers_by_key.get(event_pattern)
        if not ids:
            return
        if handler_id in ids:
            ids.remove(handler_id)
        if not ids:
            self.handlers_by_key.pop(event_pattern, None)

    @overload
    async def find(
        self,
        event_type: type[T_ExpectedEvent],
        where: None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        **event_fields: Any,
    ) -> T_ExpectedEvent | None: ...

    @overload
    async def find(
        self,
        event_type: type[T_ExpectedEvent],
        where: Callable[[T_ExpectedEvent], bool],
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        **event_fields: Any,
    ) -> T_ExpectedEvent | None: ...

    @overload
    async def find(
        self,
        event_type: PythonIdentifierStr | Literal['*'],
        where: Callable[[BaseEvent[Any]], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        **event_fields: Any,
    ) -> BaseEvent[Any] | None: ...

    async def find(
        self,
        event_type: EventPatternType,
        where: Callable[[Any], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        **event_fields: Any,
    ) -> BaseEvent[Any] | None:
        """
        Find an event matching criteria in history and/or future.

        Mirrors TS `EventBus.find` behavior:
        - Default behavior with no options: `past=True`, `future=False`
        - Search history and return the most recent match
        - Optionally wait for future emits
        - Supports exact-match equality filters via keyword args for any event field

        Args:
            event_type: The event type string or model class to find
            where: Predicate function for filtering (default: lambda _: True)
            child_of: Only match events that are descendants of this parent event
            past: Controls history search behavior:
                - True: search all history
                - False: skip history search
                - float: search events from last N seconds only
                - timedelta: search events from last N seconds
            future: Controls future wait behavior:
                - True: wait forever for matching event
                - False: don't wait for future events
                - float: wait up to N seconds for matching event
            **event_fields: Optional exact-match filters for any event field
                (for example `event_status='completed'`, `user_id='u-1'`)

        Returns:
            Matching event or None if not found/timeout
        """
        # `limit` field-equality filter would collide with filter()'s cap arg; route it through `where`.
        if 'limit' in event_fields:
            limit_field_value = event_fields.pop('limit')
            sentinel = object()
            prior_where = where

            def where_with_limit_field(event: BaseEvent[Any]) -> bool:
                return getattr(event, 'limit', sentinel) == limit_field_value and (prior_where is None or prior_where(event))

            where = where_with_limit_field
        results = await self.filter(
            event_type,
            where=where,
            child_of=child_of,
            past=past,
            future=future,
            limit=1,
            **event_fields,
        )
        return results[0] if results else None

    @overload
    async def filter(
        self,
        event_type: type[T_ExpectedEvent],
        where: None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        limit: int | None = None,
        **event_fields: Any,
    ) -> list[T_ExpectedEvent]: ...

    @overload
    async def filter(
        self,
        event_type: type[T_ExpectedEvent],
        where: Callable[[T_ExpectedEvent], bool],
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        limit: int | None = None,
        **event_fields: Any,
    ) -> list[T_ExpectedEvent]: ...

    @overload
    async def filter(
        self,
        event_type: PythonIdentifierStr | Literal['*'],
        where: Callable[[BaseEvent[Any]], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        limit: int | None = None,
        **event_fields: Any,
    ) -> list[BaseEvent[Any]]: ...

    async def filter(  # pyright: ignore[reportInconsistentOverload]
        self,
        event_type: EventPatternType,
        where: Callable[[Any], bool] | None = None,
        child_of: BaseEvent[Any] | None = None,
        past: bool | float | timedelta | None = None,
        future: bool | float | None = None,
        limit: int | None = None,
        **event_fields: Any,
    ) -> list[BaseEvent[Any]]:
        """
        Same as :meth:`find` but returns the list of all matching events
        (newest to oldest) instead of just the first match. Accepts an
        additional ``limit`` argument to cap the number of results.
        """
        self._raise_if_destroyed()
        return await self.event_history.filter(
            event_type,
            where=where,
            child_of=child_of,
            past=past,
            future=future,
            event_is_child_of=self.event_is_child_of,
            wait_for_future_match=self._wait_for_future_match,
            limit=limit,
            **event_fields,
        )

    def event_is_child_of(self, event: BaseEvent[Any], ancestor: BaseEvent[Any]) -> bool:
        """
        Check if event is a descendant of ancestor (child, grandchild, etc.).

        Walks up the parent chain from event looking for ancestor.
        Returns True if ancestor is found in the chain, False otherwise.

        Args:
            event: The potential descendant event
            ancestor: The potential ancestor event

        Returns:
            True if event is a descendant of ancestor, False otherwise
        """
        current_id = event.event_parent_id
        visited: set[str] = set()

        while current_id and current_id not in visited:
            if current_id == ancestor.event_id:
                return True
            visited.add(current_id)

            # Find parent event in any bus's history
            parent = self.event_history.get(current_id)
            if parent is None:
                # Check other buses
                for bus in list(type(self).all_instances):
                    if bus is not self and current_id in bus.event_history:
                        parent = bus.event_history[current_id]
                        break
            if parent is None:
                break
            current_id = parent.event_parent_id

        return False

    def event_is_parent_of(self, event: BaseEvent[Any], descendant: BaseEvent[Any]) -> bool:
        return self.event_is_child_of(descendant, event)

    def _start(self) -> None:
        """Start the event bus if not already running"""
        self._raise_if_destroyed()
        if not self._is_running:
            try:
                loop = asyncio.get_running_loop()

                # Hook into the event loop's close method to cleanup before it closes
                # this is necessary to silence "RuntimeError: no running event loop" and "event loop is closed" errors on shutdown
                eventbus_class = type(self)
                registered_eventbuses: weakref.WeakSet[EventBus] | None = eventbus_class._loop_eventbus_instances.get(loop)
                if registered_eventbuses is None:
                    registered_eventbuses = weakref.WeakSet()
                    eventbus_class._loop_eventbus_instances[loop] = registered_eventbuses

                if loop not in eventbus_class._loop_original_close:
                    original_close = loop.close
                    eventbus_class._loop_original_close[loop] = original_close

                    def close_with_cleanup() -> None:
                        empty_eventbuses: weakref.WeakSet[EventBus] = weakref.WeakSet()
                        # Clean up all registered EventBuses before closing the loop
                        for eventbus in list(eventbus_class._loop_eventbus_instances.get(loop, empty_eventbuses)):
                            try:
                                # Exit the EventBus loop while the asyncio loop is still running
                                if eventbus._is_running:
                                    eventbus._is_running = False

                                    # Shutdown the queue properly - our custom queue will handle cleanup
                                    if eventbus.pending_event_queue:
                                        eventbus.pending_event_queue.shutdown(immediate=True)

                                    if eventbus._runloop_task and not eventbus._runloop_task.done():
                                        # Suppress warning before cancelling
                                        setattr(eventbus._runloop_task, '_log_destroy_pending', False)
                                        eventbus._runloop_task.cancel()
                            except Exception:
                                pass

                        # Now close the loop
                        original_close()

                    setattr(loop, 'close', close_with_cleanup)

                # Register this EventBus instance in the per-loop WeakSet.
                registered_eventbuses.add(self)

                # Create async objects if needed
                if self.pending_event_queue is None:
                    # Keep queue unbounded so naive emit floods can enqueue without
                    # artificial queue caps; queue stores event object references.
                    self.pending_event_queue = CleanShutdownQueue[BaseEvent[Any]](maxsize=0)
                    self._on_idle = asyncio.Event()
                    if not self._has_inflight_events_fast() and self.pending_event_queue.qsize() == 0:
                        self._on_idle.set()
                    else:
                        self._on_idle.clear()

                # Create and start the run loop task.
                # Use a weakref-based runner so an unreferenced EventBus can be GC'd
                # without requiring explicit destroy(clear=True) by callers.
                # Run loops must start with a clean context. If emit() is called
                # from inside a handler, lock-depth ContextVars would otherwise leak
                # into the new task and bypass event lock acquisition.
                self._runloop_task = loop.create_task(
                    EventBus._run_loop_weak(weakref.ref(self)),
                    name=f'{self}._run_loop',
                    context=contextvars.Context(),
                )
                self._is_running = True
            except RuntimeError:
                # No event loop - will start when one becomes available
                pass

    async def destroy(self, clear: bool = True) -> None:
        """Destroy the event bus immediately and prevent further use

        Args:
            clear: If True, clear handlers/history for memory cleanup. If False, preserve
                handlers/history for inspection. Destroy is terminal either way.
        """
        if self._destroyed:
            if clear:
                self.event_history.clear()
                self.handlers.clear()
                self.handlers_by_key.clear()
                self.middlewares.clear()
            return

        queue_size = self.pending_event_queue.qsize() if self.pending_event_queue else 0
        has_inflight = self._has_inflight_events_fast()
        if queue_size or has_inflight:
            logger.debug(
                '⚠️ %s destroying with pending events: queue=%d inflight=%s history=%d',
                self,
                queue_size,
                has_inflight,
                len(self.event_history),
            )

        # Signal shutdown
        self._is_running = False
        self._destroyed = True

        # Shutdown the queue to unblock any pending get() operations
        if self.pending_event_queue:
            self.pending_event_queue.shutdown(immediate=True)
            self.pending_event_queue._queue.clear()  # pyright: ignore[reportPrivateUsage]

        # print('DESTROYING', self.event_history)

        if self._runloop_task and not self._runloop_task.done():
            try:
                self._runloop_task.cancel()
                self._runloop_task.add_done_callback(lambda task: task.exception() if not task.cancelled() else None)
            except Exception:
                pass

        if self._parallel_event_tasks:
            for task in list(self._parallel_event_tasks):
                if not task.done():
                    task.cancel()
            self._parallel_event_tasks.clear()

        # Clear references
        self._runloop_task = None
        self.in_flight_event_ids.clear()
        self.processing_event_ids.clear()
        self._pending_handler_changes.clear()
        for waiter in tuple(self.find_waiters):
            self.find_waiters.discard(waiter)
            if waiter.timeout_handle is not None:
                waiter.timeout_handle.cancel()
            if not waiter.future.done():
                waiter.future.set_result(None)
        if self._on_idle:
            self._on_idle.set()

        # Clear event history and handlers if requested (for memory cleanup)
        if clear:
            self.event_history.clear()
            self.handlers.clear()
            self.handlers_by_key.clear()
            self.in_flight_event_ids.clear()
            self.processing_event_ids.clear()
            self._parallel_event_tasks.clear()
            self._pending_middleware_tasks.clear()
            self.middlewares.clear()
            self.pending_event_queue = None
            self._on_idle = None
            self._lock_for_event_bus_serial = ReentrantLock()
            self.locks = LockManager()

            logger.debug('🧹 %s cleared event history and removed from global tracking', self)
        else:
            self.pending_event_queue = None
            self._on_idle = None
            self._lock_for_event_bus_serial = ReentrantLock()
            self.locks = LockManager()

        # Rename and remove from tracking even when clear=False. clear=False preserves
        # inspectable handlers/history, but destroy is terminal and non-resumable.
        self.name = f'_destroyed_{self.id[-8:]}'
        if self in type(self).all_instances:
            type(self).all_instances.discard(self)
        try:
            loop = asyncio.get_running_loop()
            registered_eventbuses = type(self)._loop_eventbus_instances.get(loop)
            if registered_eventbuses is not None:
                registered_eventbuses.discard(self)
        except RuntimeError:
            pass

        logger.debug('🛑 %s shut down immediately', self)

        # Check total memory usage across all instances
        try:
            self._check_total_memory_usage()
        except Exception:
            # Don't let memory check errors prevent shutdown
            pass

    async def wait_until_idle(self, timeout: float | None = None) -> None:
        """Wait until the event bus is idle (no events being processed and all handlers completed)"""

        self._start()
        assert self._on_idle and self.pending_event_queue, 'EventBus._start() must be called before wait_until_idle() is reached'

        start_time = asyncio.get_event_loop().time()
        remaining_timeout = timeout

        try:
            # Wait until both queue and inflight execution are empty.
            # Avoid relying on queue.join() because unfinished-task counters can
            # drift under queue-jump paths while observable runtime state is idle.
            while True:
                queue_empty = self.pending_event_queue.qsize() == 0
                has_inflight = self._has_inflight_events_fast()
                if queue_empty and not has_inflight:
                    self._on_idle.set()
                    break

                if timeout is not None:
                    elapsed = asyncio.get_event_loop().time() - start_time
                    remaining_timeout = max(0, timeout - elapsed)
                    if remaining_timeout <= 0:
                        raise TimeoutError()

                # Wait again for an idle transition.
                idle_task = asyncio.create_task(self._on_idle.wait())
                await asyncio.wait_for(idle_task, timeout=remaining_timeout)
                await asyncio.sleep(0)  # Yield again

        except TimeoutError:
            logger.warning(
                '⌛️ %s Timeout waiting for event bus to be idle after %ss (history=%d)',
                self,
                timeout,
                len(self.event_history),
            )

    @staticmethod
    async def _run_loop_weak(bus_ref: 'weakref.ReferenceType[EventBus]') -> None:
        """
        Weakref-based run loop.

        This runner avoids holding a strong EventBus reference while idle,
        allowing unreferenced buses to be garbage-collected naturally without
        an explicit destroy().
        """
        try:
            while True:
                bus = bus_ref()
                if bus is None or not bus._is_running:
                    break

                queue = bus.pending_event_queue
                on_idle = bus._on_idle
                del bus

                if queue is None or on_idle is None:
                    await asyncio.sleep(0.01)
                    continue

                event: BaseEvent[Any] | None = None
                try:
                    event = await queue.get()
                except QueueShutDown:
                    break
                except asyncio.CancelledError:
                    break
                except RuntimeError as e:
                    if 'Event loop is closed' in str(e) or 'no running event loop' in str(e):
                        break
                    logger.exception(f'❌ Weak run loop runtime error: {type(e).__name__} {e}', exc_info=True)
                    continue
                except Exception as e:
                    logger.exception(f'❌ Weak run loop error: {type(e).__name__} {e}', exc_info=True)
                    continue

                bus = bus_ref()
                if bus is None:
                    try:
                        queue.task_done()
                    except Exception:
                        pass
                    break

                try:
                    if bus._on_idle:
                        bus._on_idle.clear()

                    bus.processing_event_ids.add(event.event_id)

                    event_lock = bus.locks.get_lock_for_event(bus, event)
                    if event_lock is None:

                        async def process_parallel_event(
                            bus: 'EventBus' = bus,
                            event: BaseEvent[Any] = event,
                            queue: CleanShutdownQueue[BaseEvent[Any]] = queue,
                        ) -> None:
                            try:
                                await bus.step(event=event)
                            finally:
                                try:
                                    queue.task_done()
                                except ValueError:
                                    pass

                        task = asyncio.create_task(
                            process_parallel_event(),
                            name=f'{bus}._process_event({event.event_id[-4:]})',
                        )
                        bus._parallel_event_tasks.add(task)

                        def _on_done(
                            done_task: asyncio.Task[None], *, bus_ref: 'weakref.ReferenceType[EventBus]' = weakref.ref(bus)
                        ):
                            live_bus = bus_ref()
                            if live_bus is not None:
                                live_bus._parallel_event_tasks.discard(done_task)
                                if (
                                    live_bus._on_idle
                                    and live_bus.pending_event_queue
                                    and not live_bus._has_inflight_events_fast()
                                    and live_bus.pending_event_queue.qsize() == 0
                                ):
                                    live_bus._on_idle.set()
                            if done_task.cancelled():
                                return
                            try:
                                exc = done_task.exception()
                            except asyncio.CancelledError:
                                return
                            if exc is not None:
                                logger.exception('❌ Weak run loop parallel event task error: %s %s', type(exc).__name__, exc)

                        task.add_done_callback(_on_done)
                    else:
                        try:
                            await bus.step(event=event)
                        finally:
                            try:
                                queue.task_done()
                            except ValueError:
                                pass
                        if bus._on_idle and bus.pending_event_queue:
                            if not bus._has_inflight_events_fast() and bus.pending_event_queue.qsize() == 0:
                                bus._on_idle.set()
                except QueueShutDown:
                    break
                except asyncio.CancelledError:
                    break
                except RuntimeError as e:
                    if 'Event loop is closed' in str(e) or 'no running event loop' in str(e):
                        break
                    logger.exception(f'❌ Weak run loop runtime error: {type(e).__name__} {e}', exc_info=True)
                except Exception as e:
                    logger.exception(f'❌ Weak run loop error: {type(e).__name__} {e}', exc_info=True)
                finally:
                    del bus
        finally:
            bus = bus_ref()
            if bus is not None:
                bus._is_running = False

    async def _get_next_event(self, wait_for_timeout: float = 0.1) -> 'BaseEvent[Any] | None':
        """Get the next event from the queue"""

        assert self._on_idle and self.pending_event_queue, 'EventBus._start() must be called before _get_next_event()'
        if not self._is_running:
            return None

        try:
            # Create a task for queue.get() so we can cancel it cleanly
            get_next_queued_event = asyncio.create_task(self.pending_event_queue.get())
            setattr(get_next_queued_event, '_log_destroy_pending', False)  # Suppress warnings on cleanup

            # Wait for next event with timeout
            has_next_event, _pending = await asyncio.wait({get_next_queued_event}, timeout=wait_for_timeout)
            if has_next_event:
                # Check if we're still running before returning the event
                if not self._is_running:
                    get_next_queued_event.cancel()
                    return None
                return await get_next_queued_event  # await to actually resolve it to the next event
            else:
                # Get task timed out, cancel it cleanly to suppress warnings
                get_next_queued_event.cancel()

                # Check if we're idle, if so, set the idle flag
                if not self._has_inflight_events_fast() and self.pending_event_queue.qsize() == 0:
                    self._on_idle.set()
                return None

        except (asyncio.CancelledError, RuntimeError, QueueShutDown):
            # Clean cancellation during shutdown or queue was shut down
            return None

    async def _finalize_local_event_processing(self, event: BaseEvent[Any]) -> None:
        """
        Clear local in-flight markers and run completion propagation exactly once.

        This is shared by both `step()` and the weak runloop path so completion
        semantics stay identical regardless of which runner consumed the event.
        """
        self.processing_event_ids.discard(event.event_id)
        # Local bus consumed this event instance (or observed completion), so it
        # should not remain in this bus's active set.
        self.in_flight_event_ids.discard(event.event_id)

        newly_completed_events = self._mark_event_tree_complete_if_ready(event)
        for completed_event in newly_completed_events:
            await self.on_event_change(completed_event, EventStatus.COMPLETED)

    def _mark_event_tree_complete_if_ready(self, root_event: BaseEvent[Any]) -> list[BaseEvent[Any]]:
        """
        Re-check completion for `root_event` and descendants in post-order.

        Timeout/cancellation paths can update child result statuses after an
        earlier completion check. Running this post-order pass ensures children
        are marked complete before their parents are re-evaluated.
        """
        newly_completed: list[BaseEvent[Any]] = []
        visited_event_ids: set[str] = set()

        def visit(event: BaseEvent[Any]) -> None:
            if event.event_id in visited_event_ids:
                return
            visited_event_ids.add(event.event_id)

            for child_event in event.event_children:
                visit(child_event)

            was_complete = event.event_status == EventStatus.COMPLETED
            # Only the root event may still appear "in-flight" on this bus during finalization.
            # Descendants are not currently being processed in this frame, so they must consider
            # queues on this bus too (otherwise queued children can be marked complete too early).
            current_bus = self if event.event_id == root_event.event_id else None
            event._mark_completed(current_bus=current_bus)  # pyright: ignore[reportPrivateUsage]
            just_completed = (not was_complete) and event.event_status == EventStatus.COMPLETED
            if just_completed:
                self._mark_event_complete_on_all_buses(event)
                newly_completed.append(event)

        visit(root_event)
        return newly_completed

    async def step(
        self, event: 'BaseEvent[Any] | None' = None, timeout: float | None = None, wait_for_timeout: float = 0.1
    ) -> 'BaseEvent[Any] | None':
        """
        Consume and process a single event from the queue (one iteration of the run loop).

        This is the high-level "consumer" method that:
        1. Dequeues the next event (or uses one passed in)
        2. Acquires the event lock selected by concurrency mode
        3. Calls _process_event() to execute handlers
        4. Marks the queue task as done (only if event came from queue)
        5. Manages idle state signaling

        Use this method when manually driving the event loop (e.g., in tests).
        For automatic processing, use emit() which queues events for the run loop.

        Args:
            event: Optional event to process directly (bypasses queue if provided)
            timeout: Handler execution timeout in seconds
            wait_for_timeout: How long to wait for next event from queue (default: 0.1s)

        Returns:
            The processed event, or None if queue was empty/shutdown

        Warning:
            Passing an event directly (bypassing the queue) is for advanced use only, be aware if:

            - **Event not in queue**: Works fine, handlers execute normally.
            - **Event already completed**: Handlers will run AGAIN, overwriting previous
              results. No guard against double-processing.
            - **Event in queue but not next**: Event processes immediately, but STAYS
              in queue. The run loop will process it again later (double-processing).

        See Also:
            emit: Queues an event for normal async processing by the bus's existing run loop (recommended)
            _process_event: Lower-level method that executes handlers (called by step)
        """
        assert self._on_idle and self.pending_event_queue, 'EventBus._start() must be called before step()'

        # Track if we got the event from the queue
        from_queue = False

        # Wait for next event with timeout to periodically check idle state
        if event is None:
            event = await self._get_next_event(wait_for_timeout=wait_for_timeout)
            from_queue = True
        if event is None:
            return None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('🏃 %s.step(%s) STARTING', self, event)

        # Lifecycle note:
        # - Every `step()` transition starts in "not idle" state.
        # - The caller may be the runloop (`from_queue=True`) or a direct queue-jump
        #   call (`from_queue=False`) from `BaseEvent._process_self_on_all_buses()`.
        # - Idle can be restored only after queue bookkeeping and in-flight markers
        #   are both fully reconciled in `finally`.
        self._on_idle.clear()

        # Acquire the event lock selected by event/bus concurrency policy.
        self.processing_event_ids.add(event.event_id)
        try:
            async with self.locks._run_with_event_lock(self, event):  # pyright: ignore[reportPrivateUsage]
                # Process the event
                if event.event_status != EventStatus.COMPLETED:
                    await self._process_event(event, timeout=timeout)

                # Queue lifecycle:
                # - `queue.get()` increments `_unfinished_tasks`.
                # - We must call `task_done()` exactly once for that consume path.
                # - Direct `step(event=...)` calls bypass `queue.get()` and therefore
                #   must not call `task_done()`.
                if from_queue:
                    self.pending_event_queue.task_done()
        finally:
            await self._finalize_local_event_processing(event)
            # Idle lifecycle reconciliation:
            # - The runloop normally restores `_on_idle` after each queue turn.
            # - Direct `step(event=...)` calls have no subsequent runloop turn, so if
            #   this step drained the last in-flight work, `_on_idle` could remain
            #   permanently cleared and `wait_until_idle()` would block forever.
            # - Setting `_on_idle` here when queue+inflight are both empty keeps idle
            #   semantics identical for runloop and direct-step execution paths.
            if self._on_idle and self.pending_event_queue:
                if not self._has_inflight_events_fast() and self.pending_event_queue.qsize() == 0:
                    self._on_idle.set()

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('✅ %s.step(%s) COMPLETE', self, event)
        return event

    def _create_slow_event_warning_timer(
        self,
        event: BaseEvent[Any],
    ) -> Callable[[], Coroutine[Any, Any, None]] | None:
        event_slow_timeout = self._resolve_event_slow_timeout(event, self)
        if event_slow_timeout is None or event_slow_timeout <= 0:
            return None
        return partial(self._slow_event_warning_monitor, event, event_slow_timeout)

    async def _mark_event_complete_if_ready(self, event: BaseEvent[Any]) -> None:
        was_complete = event.event_status == EventStatus.COMPLETED
        event._mark_completed(current_bus=self)  # pyright: ignore[reportPrivateUsage]
        just_completed = (not was_complete) and event.event_status == EventStatus.COMPLETED
        if just_completed:
            self._mark_event_complete_on_all_buses(event)
            await self.on_event_change(event, EventStatus.COMPLETED)

    async def _propagate_parent_completion(self, event: BaseEvent[Any]) -> None:
        current = event
        checked_ids: set[str] = set()

        while current.event_parent_id and current.event_parent_id not in checked_ids:
            checked_ids.add(current.event_parent_id)

            parent_event = None
            parent_bus: EventBus | None = None
            for bus in list(type(self).all_instances):
                if bus and current.event_parent_id in bus.event_history:
                    parent_event = bus.event_history[current.event_parent_id]
                    parent_bus = bus
                    break

            if not parent_event:
                break

            was_complete = parent_event.event_status == EventStatus.COMPLETED
            if not was_complete:
                parent_event._mark_completed(current_bus=parent_bus)  # pyright: ignore[reportPrivateUsage]
            just_completed = (not was_complete) and parent_event.event_status == EventStatus.COMPLETED
            if parent_bus and just_completed:
                self._mark_event_complete_on_all_buses(parent_event)
                await parent_bus.on_event_change(parent_event, EventStatus.COMPLETED)

            current = parent_event

    def _trim_event_history_if_needed(self) -> None:
        if (
            self.event_history.max_history_size is not None
            and self.event_history.max_history_size > 0
            and self.event_history.max_history_drop
            and len(self.event_history) > self.event_history.max_history_size
        ):
            self.event_history.trim_event_history(owner_label=str(self))

    async def _process_event(self, event: BaseEvent[Any], timeout: float | None = None) -> None:
        """
        Execute all applicable handlers for an event (low-level, assumes lock is held).

        This is the core event handling method that:
        1. Finds all applicable handlers (type-specific + wildcard)
        2. Creates pending EventResult placeholders
        3. Executes handlers (serially or in parallel based on bus config)
        4. Marks the event as complete when all handlers finish
        5. Propagates completion status up the parent event chain
        6. Cleans up event history if over size limit

        IMPORTANT: This method assumes the caller already applied `locks._run_with_event_lock(...)`
        for the event execution.
        For safe external use, call step() instead which handles locking.

        Args:
            event: The event to handle
            timeout: Handler execution timeout in seconds (defaults to event.event_timeout)

        Warning:
            This is a low-level method with no safety guards. Behavior in edge cases:

            - **Event not in queue**: Works fine, handlers execute normally. This method
              does not interact with the queue at all.
            - **Event already completed**: Handlers run AGAIN, ``_create_pending_handler_results()``
              overwrites previous results. No guard against double-processing.
            - **Event in queue but not next**: Works fine for this call, but event stays
              in queue and will be processed again later by the run loop.
            - **Another event being processed (lock held elsewhere)**: If called without
              holding the lock, concurrent handler execution may cause race conditions.
              If called from within a handler (lock is re-entrant), causes nested processing.
            - **This exact event already being processed**: Recursive/re-entrant processing.
              Handlers run again while already running, results overwritten mid-execution.
              Likely to cause undefined behavior.

        See Also:
            step: High-level method that acquires lock and calls _process_event
            emit: Queues an event for async processing (recommended)
        """
        # Get applicable handlers
        applicable_handlers = self._get_handlers_for_event(event)
        slow_event_monitor_factory = self._create_slow_event_warning_timer(event)
        resolved_event_timeout = timeout if timeout is not None else event.event_timeout
        if resolved_event_timeout is None:
            resolved_event_timeout = self.event_timeout
        if resolved_event_timeout is not None and resolved_event_timeout <= 0:
            resolved_event_timeout = None

        await self.on_event_change(event, EventStatus.PENDING)

        # Execute handlers
        timeout_scope = asyncio.timeout(resolved_event_timeout)
        try:
            async with timeout_scope:
                async with _run_with_slow_monitor(
                    slow_event_monitor_factory,
                    task_name=f'{self}.slow_event_monitor({event})',
                ):
                    await event._run_handlers(  # pyright: ignore[reportPrivateUsage]
                        eventbus=self,
                        handlers=applicable_handlers,
                        timeout=resolved_event_timeout,
                    )
        except TimeoutError:
            if not timeout_scope.expired():
                raise
            assert resolved_event_timeout is not None
            await self._finalize_event_timeout(event, resolved_event_timeout)

        await self._mark_event_complete_if_ready(event)
        await self._propagate_parent_completion(event)
        self._trim_event_history_if_needed()

    def _get_handlers_for_event(self, event: BaseEvent[Any]) -> dict[PythonIdStr, EventHandler]:
        """Get all handlers that should process the given event, filtering out those that would create loops"""
        applicable_handlers: list[EventHandler] = []

        for key in (event.event_type, '*'):
            indexed_ids = self.handlers_by_key.get(key, [])
            if not indexed_ids:
                continue
            for handler_id in indexed_ids:
                handler_entry = self.handlers.get(handler_id)
                if handler_entry:
                    applicable_handlers.append(handler_entry)

        # Filter out handlers that would create loops and build id->handler mapping
        # Use handler id as key to preserve all handlers even with duplicate names
        filtered_handlers: dict[PythonIdStr, EventHandler] = {}
        for handler_entry in applicable_handlers:
            if self._would_create_loop(event, handler_entry):
                continue
            else:
                assert handler_entry.id is not None
                filtered_handlers[handler_entry.id] = handler_entry
                # logger.debug(f'  Found handler {handler_entry.handler_name}#{handler_entry.id[-4:]}()')

        return filtered_handlers

    @contextmanager
    def _run_with_handler_dispatch_context(self, event: BaseEvent[T_EventResultType], handler_id: str):
        """Scope ContextVar state for one handler execution.

        This is the single handler execution context manager used by
        ``EventResult.run_handler(...)`` for both sync and async handlers. It sets
        current event/handler/bus ContextVars and mirrors event-lock ownership
        into copied dispatch contexts via ``locks._run_with_handler_dispatch_context(...)``.
        """
        event_token = EventBus.current_event_context.set(event)
        current_handler_token = EventBus.current_handler_id_context.set(handler_id)
        current_eventbus_token = EventBus.current_eventbus_context.set(self)
        try:
            with self.locks._run_with_handler_dispatch_context(self, event):  # pyright: ignore[reportPrivateUsage]
                yield
        finally:
            EventBus.current_event_context.reset(event_token)
            EventBus.current_handler_id_context.reset(current_handler_token)
            EventBus.current_eventbus_context.reset(current_eventbus_token)

    async def _finalize_event_timeout(self, event: BaseEvent[Any], timeout_seconds: float) -> None:
        """Finalize event-level hard timeout across pending/started handler results.

        - pending results become ``EventHandlerCancelledError``
        - started results become ``EventHandlerAbortedError``
        - child event processing is cancelled through event-level propagation
        """
        timeout_error = TimeoutError(
            f'Event {self.label}.on({event.event_type}#{event.event_id[-4:]}) timed out after {timeout_seconds}s'
        )
        event._cancel_pending_child_processing(timeout_error)  # pyright: ignore[reportPrivateUsage]
        for event_result in event.event_results.values():
            if event_result.status == 'pending':
                event_result.update(
                    error=EventHandlerCancelledError(
                        f'Cancelled: event {event.event_type}#{event.event_id[-4:]} timed out after {timeout_seconds}s'
                    )
                )
                await self.on_event_result_change(event, event_result, EventStatus.COMPLETED)
            elif event_result.status == 'started':
                event_result.update(
                    error=EventHandlerAbortedError(
                        f'Event handler {event_result.handler.label}({event}) was interrupted because the event timed out after {timeout_seconds}s'
                    )
                )
                await self.on_event_result_change(event, event_result, EventStatus.COMPLETED)

    async def _run_handler(
        self,
        event: 'BaseEvent[T_EventResultType]',
        handler_entry: EventHandler,
        timeout: float | None = None,
    ) -> T_EventResultType | BaseEvent[Any] | None:
        """Safely execute a single handler with middleware support and EventResult orchestration."""

        handler_id = handler_entry.id
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                ' ↳ %s._run_handler(%s, handler=%s#%s)',
                self,
                event,
                handler_entry.handler_name,
                handler_id[-4:],
            )

        resolved_timeout = self._resolve_handler_timeout(event, handler_entry, self, timeout_override=timeout)
        resolved_slow_timeout = self._resolve_handler_slow_timeout(event, handler_entry, self)

        if handler_id not in event.event_results:
            new_results = event._create_pending_handler_results(  # pyright: ignore[reportPrivateUsage]
                {handler_id: handler_entry}, eventbus=self, timeout=resolved_timeout
            )
            for pending_result in new_results.values():
                await self.on_event_result_change(event, pending_result, EventStatus.PENDING)

        first_handler_id = next(iter(event.event_results), None)
        event_result = event.event_results[handler_id]

        try:
            result_value = await event_result.run_handler(
                event,
                eventbus=self,
                timeout=resolved_timeout,
                handler_slow_timeout=resolved_slow_timeout,
                notify_event_started=(first_handler_id == handler_id),
                format_exception_for_log=log_filtered_traceback,
            )

            result_type_name = type(result_value).__name__ if result_value is not None else 'None'
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    '    ↳ Handler %s#%s returned: %s',
                    handler_entry.handler_name,
                    handler_id[-4:],
                    result_type_name,
                )

            await self.on_event_result_change(event, event_result, EventStatus.COMPLETED)
            return result_value

        except asyncio.CancelledError:
            await self.on_event_result_change(event, event_result, EventStatus.COMPLETED)
            raise
        except Exception:
            await self.on_event_result_change(event, event_result, EventStatus.COMPLETED)
            raise

    def _would_create_loop(self, event: BaseEvent[Any], handler_entry: EventHandler) -> bool:
        """Check if calling this handler would create a loop"""
        handler = handler_entry.handler
        if handler is None:
            return False

        # First check: If handler is another EventBus emit/dispatch method, check if we're forwarding to another bus that it's already been processed by
        if inspect.ismethod(handler):
            bound_self = handler.__self__
            bound_name = handler.__name__
            if isinstance(bound_self, EventBus) and bound_name in ('emit', 'dispatch'):
                target_bus = bound_self
                if target_bus.label in event.event_path:
                    logger.debug(
                        f'⚠️ {self} handler {handler_entry.label}({event}) skipped to prevent infinite forwarding loop with {target_bus.label}'
                    )
                    return True

        # Second check: if this handler already has an in-flight/completed result for this
        # event on this bus, avoid re-entrancy. Rehydrated events can legitimately contain
        # pending placeholders from a previous process; those must remain runnable.
        handler_id = handler_entry.id
        if handler_id in event.event_results:
            existing_result = event.event_results[handler_id]
            if existing_result.status == 'started':
                logger.debug(
                    f'⚠️ {self} handler {handler_entry.label}({event}) is already {existing_result.status} for event {event.event_id} (preventing recursive call)'
                )
                return True
            if existing_result.status == 'pending' and event.event_status == EventStatus.STARTED:
                logger.debug(
                    f'⚠️ {self} handler {handler_entry.label}({event}) is already pending while event is started for event {event.event_id} (preventing recursive call)'
                )
                return True
            if existing_result.status in ('completed', 'error') or existing_result.completed_at is not None:
                logger.debug(
                    f'⚠️ {self} handler {handler_entry.label}({event}) already completed @ {existing_result.completed_at} for event {event.event_id} (will not re-run)'
                )
                return True

        # Third check: For non-forwarding handlers, check recursion depth
        # Forwarding handlers (EventBus.emit / EventBus.dispatch) are allowed to forward at any depth
        is_forwarding_handler = (
            inspect.ismethod(handler) and isinstance(handler.__self__, EventBus) and handler.__name__ in ('emit', 'dispatch')
        )

        if not is_forwarding_handler:
            # Only check recursion for regular handlers, not forwarding
            recursion_depth = self._handler_dispatched_ancestor(event, handler_id)
            if recursion_depth > self.max_handler_recursion_depth:
                raise RuntimeError(
                    f'Infinite loop detected: Handler {handler_entry.label} '
                    f'has recursively processed {recursion_depth} levels of events. '
                    f'Current event: {event}, Handler: {handler_id}'
                )
            elif recursion_depth == self.max_handler_recursion_depth:
                logger.warning(
                    f'⚠️ {self} handler {handler_entry.label} '
                    f'at maximum recursion depth ({self.max_handler_recursion_depth} levels) - next level will raise exception'
                )

        return False

    def _handler_dispatched_ancestor(
        self, event: BaseEvent[Any], handler_id: str, visited: set[str] | None = None, depth: int = 0
    ) -> int:
        """Check how many times this handler appears in the ancestry chain. Returns the depth count."""
        # Prevent infinite recursion in case of circular parent references
        if visited is None:
            visited = set()
        if event.event_id in visited:
            return depth
        visited.add(event.event_id)

        # If this event has no parent, it's a root event - no ancestry to check
        if not event.event_parent_id:
            return depth

        # Find parent event in any bus's history
        parent_event = None
        # Create a list copy to avoid "Set changed size during iteration" error
        for bus in list(type(self).all_instances):
            if event.event_parent_id in bus.event_history:
                parent_event = bus.event_history[event.event_parent_id]
                break

        if not parent_event:
            return depth

        # Check if this handler processed the parent event
        if handler_id in parent_event.event_results:
            result = parent_event.event_results[handler_id]
            if result.status in ('pending', 'started', 'completed'):
                # This handler processed the parent event, increment depth
                depth += 1

        # Recursively check the parent's ancestry
        return self._handler_dispatched_ancestor(parent_event, handler_id, visited, depth)

    def log_tree(self) -> str:
        """Print a nice pretty formatted tree view of all events in the history including their results and child events recursively"""
        from abxbus.logging import log_eventbus_tree

        return log_eventbus_tree(self)

    def _check_total_memory_usage(self) -> None:
        """Check total memory usage across all EventBus instances and warn if >50MB"""
        import sys

        total_bytes = 0
        bus_details: list[tuple[str, int, int, int]] = []

        # Iterate through all EventBus instances
        # Create a list copy to avoid "Set changed size during iteration" error
        for bus in list(type(self).all_instances):
            try:
                bus_bytes = 0

                # Count events in history
                for event in bus.event_history.values():
                    bus_bytes += sys.getsizeof(event)
                    # Also count the event's data
                    for attr_value in event.__dict__.values():
                        if isinstance(attr_value, (str, bytes, list, dict)):
                            bus_bytes += sys.getsizeof(_as_any(attr_value))

                # Count events in queue
                if bus.pending_event_queue:
                    for event in bus.pending_event_queue.iter_items():
                        bus_bytes += sys.getsizeof(event)
                        for attr_value in event.__dict__.values():
                            if isinstance(attr_value, (str, bytes, list, dict)):
                                bus_bytes += sys.getsizeof(_as_any(attr_value))

                total_bytes += bus_bytes
                bus_details.append(
                    (
                        bus.label,
                        bus_bytes,
                        len(bus.event_history),
                        bus.pending_event_queue.qsize() if bus.pending_event_queue else 0,
                    )
                )
            except Exception:
                # Skip buses that can't be measured
                continue

        total_mb = total_bytes / (1024 * 1024)

        if total_mb > 50:
            # Build detailed breakdown
            details: list[str] = []
            for name, bytes_used, history_size, queue_size in sorted(bus_details, key=lambda x: x[1], reverse=True):
                mb = bytes_used / (1024 * 1024)
                if mb > 0.1:  # Only show buses using >0.1MB
                    details.append(f'  - {name}: {mb:.1f}MB (history={history_size}, queue={queue_size})')

            warning_msg = (
                f'\n⚠️  WARNING: Total EventBus memory usage is {total_mb:.1f}MB (>50MB limit)\n'
                f'Active EventBus instances: {len(type(self).all_instances)}\n'
            )
            if details:
                warning_msg += 'Memory breakdown:\n' + '\n'.join(details[:5])  # Show top 5
                if len(details) > 5:
                    warning_msg += f'\n  ... and {len(details) - 5} more'

            warning_msg += '\nConsider:\n'
            warning_msg += '  - Reducing max_history_size\n'
            warning_msg += '  - Clearing completed EventBus instances with destroy(clear=True)\n'
            warning_msg += '  - Reducing event payload sizes\n'

            logger.warning(warning_msg)
